# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rqalpha.interface import AbstractMod
from rqalpha.events import EVENT
from rqalpha.utils import get_account_type
from rqalpha.const import ACCOUNT_TYPE

from .frontend_validator import StockFrontendValidator, FutureFrontendValidator


class RiskManagerMod(AbstractMod):

    def __init__(self):
        self._env = None
        self.mod_config = None
        self._frontend_validator = {}

    def start_up(self, env, mod_config):
        self._env = env
        self.mod_config = mod_config
        self._env.event_bus.prepend_listener(EVENT.ORDER_PENDING_NEW, self._frontend_validate)

    def tear_down(self, code, exception=None):
        pass

    def _frontend_validate(self, account, order):
        frontend_validator = self._get_frontend_validator_for(order.order_book_id)
        frontend_validator.order_pipeline(account, order)

    def _get_frontend_validator_for(self, order_book_id):
        account_type = get_account_type(order_book_id)
        try:
            return self._frontend_validator[account_type]
        except KeyError:
            if account_type == ACCOUNT_TYPE.STOCK:
                validator = StockFrontendValidator(self.mod_config)
            elif account_type == ACCOUNT_TYPE.FUTURE:
                validator = FutureFrontendValidator(self.mod_config)
            else:
                raise RuntimeError('account type {} not supported yet'.format(account_type))
            self._frontend_validator[account_type] = validator
            return validator




# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
from collections import defaultdict

import six
from enum import Enum
import numpy as np
import pandas as pd

from rqalpha.interface import AbstractMod
from rqalpha.events import EVENT
from rqalpha.const import ACCOUNT_TYPE, EXIT_CODE
from rqalpha.utils.risk import Risk
from rqalpha.utils.repr import properties
from rqalpha.execution_context import ExecutionContext


class AnalyserMod(AbstractMod):
    def __init__(self):
        self._env = None
        self._mod_config = None
        self._enabled = False
        self._result = None

        self._orders = defaultdict(list)
        self._trades = []
        self._total_portfolios = []
        self._sub_portfolios = defaultdict(list)
        self._positions = defaultdict(list)

        self._benchmark_daily_returns = []
        self._portfolio_daily_returns = []
        self._latest_portfolio = None
        self._latest_benchmark_portfolio = None

    def start_up(self, env, mod_config):
        self._env = env
        self._mod_config = mod_config
        self._enabled = (self._mod_config.record or self._mod_config.plot or self._mod_config.output_file or
                         self._mod_config.plot_save_file or self._mod_config.report_save_path)

        if self._enabled:
            env.event_bus.add_listener(EVENT.POST_SETTLEMENT, self._collect_daily)
            env.event_bus.add_listener(EVENT.TRADE, self._collect_trade)
            env.event_bus.add_listener(EVENT.ORDER_CREATION_PASS, self._collect_order)

    def _collect_trade(self, account, trade):
        self._trades.append(self._to_trade_record(trade))

    def _collect_order(self, account, order):
        self._orders[order.trading_datetime.date()].append(order)

    def _collect_daily(self):
        date = self._env.calendar_dt.date()
        portfolio = self._env.account.get_portfolio(date)

        self._latest_portfolio = portfolio
        self._portfolio_daily_returns.append(portfolio.daily_returns)
        self._total_portfolios.append(self._to_portfolio_record(date, portfolio))

        if ACCOUNT_TYPE.BENCHMARK in self._env.accounts:
            self._latest_benchmark_portfolio = self._env.accounts[ACCOUNT_TYPE.BENCHMARK].portfolio
            self._benchmark_daily_returns.append(self._latest_benchmark_portfolio.daily_returns)
        else:
            self._benchmark_daily_returns.append(0)

        for account_type, account in six.iteritems(self._env.accounts):
            portfolio = account.get_portfolio(date)
            self._sub_portfolios[account_type].append(self._to_portfolio_record2(date, portfolio))
            for order_book_id, position in six.iteritems(portfolio.positions):
                self._positions[account_type].append(self._to_position_record(date, order_book_id, position))

    def _symbol(self, order_book_id):
        return self._env.data_proxy.instruments(order_book_id).symbol

    @staticmethod
    def _safe_convert(value, ndigits=3):
        if isinstance(value, Enum):
            return value.name

        if isinstance(value, (float, np.float64, np.float32, np.float16, np.float)):
            return round(value, ndigits)

        return value

    def _to_portfolio_record(self, date, portfolio):
        data = {
            k: self._safe_convert(v, 3) for k, v in six.iteritems(properties(portfolio))
            if not k.startswith('_') and not k.endswith('_') and k not in {
                "positions", "start_date", "starting_cash"
            }
        }
        data['date'] = date
        return data

    def _to_portfolio_record2(self, date, portfolio):
        data = {
            k: self._safe_convert(v, 3) for k, v in six.iteritems(portfolio.__dict__)
            if not k.startswith('_') and not k.endswith('_') and k not in {
                "positions", "start_date", "starting_cash"
            }
            }
        data['date'] = date
        return data

    def _to_position_record(self, date, order_book_id, position):
        data = {
            k: self._safe_convert(v, 3) for k, v in six.iteritems(position.__dict__)
            if not k.startswith('_') and not k.endswith('_')
            }
        data['order_book_id'] = order_book_id
        data['symbol'] = self._symbol(order_book_id)
        data['date'] = date
        return data

    def _to_trade_record(self, trade):
        data = {
            k: self._safe_convert(v) for k, v in six.iteritems(properties(trade))
            if not k.startswith('_') and not k.endswith('_') and k != 'order'
        }
        data['order_book_id'] = trade.order.order_book_id
        data['symbol'] = self._symbol(trade.order.order_book_id)
        data['side'] = self._safe_convert(trade.order.side)
        data['position_effect'] = self._safe_convert(trade.order.position_effect)
        data['datetime'] = data['datetime'].strftime("%Y-%m-%d %H:%M:%S")
        data['trading_datetime'] = data['trading_datetime'].strftime("%Y-%m-%d %H:%M:%S")
        return data

    def tear_down(self, code, exception=None):
        if code != EXIT_CODE.EXIT_SUCCESS or not self._enabled:
            return

        strategy_name = os.path.basename(self._env.config.base.strategy_file).split(".")[0]
        data_proxy = self._env.data_proxy

        summary = {
            'strategy_name': strategy_name,
        }
        for k, v in six.iteritems(self._env.config.base.__dict__):
            if k in ["trading_calendar", "account_list", "timezone", "persist_mode",
                     "resume_mode", "data_bundle_path", "handle_split", "persist"]:
                continue
            summary[k] = self._safe_convert(v, 2)

        risk = Risk(np.array(self._portfolio_daily_returns), np.array(self._benchmark_daily_returns),
                    data_proxy.get_risk_free_rate(self._env.config.base.start_date, self._env.config.base.end_date),
                    (self._env.config.base.end_date - self._env.config.base.start_date).days + 1)
        summary.update({
            'alpha': self._safe_convert(risk.alpha, 3),
            'beta': self._safe_convert(risk.beta, 3),
            'sharpe': self._safe_convert(risk.sharpe, 3),
            'information_ratio': self._safe_convert(risk.information_ratio, 3),
            'downside_risk': self._safe_convert(risk.annual_downside_risk, 3),
            'tracking_error': self._safe_convert(risk.annual_tracking_error, 3),
            'sortino': self._safe_convert(risk.sortino, 3),
            'volatility': self._safe_convert(risk.annual_volatility, 3),
            'max_drawdown': self._safe_convert(risk.max_drawdown, 3),
        })

        summary.update({
            k: self._safe_convert(v, 3) for k, v in six.iteritems(properties(self._latest_portfolio))
            if k not in ["positions", "daily_returns", "daily_pnl"]
        })

        if self._latest_benchmark_portfolio:
            summary['benchmark_total_returns'] = self._latest_benchmark_portfolio.total_returns
            summary['benchmark_annualized_returns'] = self._latest_benchmark_portfolio.annualized_returns

        trades = pd.DataFrame(self._trades)
        if 'datetime' in trades.columns:
            trades = trades.set_index('datetime')

        df = pd.DataFrame(self._total_portfolios)
        df['date'] = pd.to_datetime(df['date'])
        total_portfolios = df.set_index('date').sort_index()

        result_dict = {
            'summary': summary,
            'trades': trades,
            'total_portfolios': total_portfolios,
        }

        if ExecutionContext.plots is not None:
            plots = ExecutionContext.plots.get_plots()
            plots_items = defaultdict(dict)
            for series_name, value_dict in six.iteritems(plots):
                for date, value in six.iteritems(value_dict):
                    plots_items[date][series_name] = value
                    plots_items[date]["date"] = date

            df = pd.DataFrame([dict_data for date, dict_data in six.iteritems(plots_items)])
            df["date"] = pd.to_datetime(df["date"])
            df = df.set_index("date").sort_index()
            result_dict["plots"] = df

        for account_type, account in six.iteritems(self._env.accounts):
            account_name = account_type.name.lower()
            portfolios_list = self._sub_portfolios[account_type]
            df = pd.DataFrame(portfolios_list)
            df["date"] = pd.to_datetime(df["date"])
            portfolios_df = df.set_index("date").sort_index()
            result_dict["{}_portfolios".format(account_name)] = portfolios_df

            positions_list = self._positions[account_type]
            positions_df = pd.DataFrame(positions_list)
            if "date" in positions_df.columns:
                positions_df["date"] = pd.to_datetime(positions_df["date"])
                positions_df = positions_df.set_index("date").sort_index()
            result_dict["{}_positions".format(account_name)] = positions_df

        self._result = result_dict

        if self._mod_config.output_file:
            with open(self._mod_config.output_file, 'wb') as f:
                pickle.dump(result_dict, f)

        if self._mod_config.plot:
            from rqalpha.plot import plot_result
            plot_result(result_dict)

        if self._mod_config.plot_save_file:
            from rqalpha.plot import plot_result
            plot_result(result_dict, False, self._mod_config.plot_save_file)

        if self._mod_config.report_save_path:
            from rqalpha.utils.report import generate_report
            generate_report(result_dict, self._mod_config.report_save_path)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import pandas as pd

from .base_account import BaseAccount
from ..dividend import Dividend
from ...const import SIDE, ACCOUNT_TYPE
from ...utils.i18n import gettext as _
from ...utils.logger import user_system_log, system_log
from ...execution_context import ExecutionContext


class StockAccount(BaseAccount):
    def __init__(self, env, init_cash, start_date):
        super(StockAccount, self).__init__(env, init_cash, start_date, ACCOUNT_TYPE.STOCK)

    def before_trading(self):
        super(StockAccount, self).before_trading()
        positions = self.portfolio.positions
        removing_ids = []
        for order_book_id in positions.keys():
            position = positions[order_book_id]
            if position._quantity == 0:
                removing_ids.append(order_book_id)
        for order_book_id in removing_ids:
            positions.pop(order_book_id, None)
        trading_date = ExecutionContext.get_current_trading_dt().date()
        self._handle_dividend_payable(trading_date)
        if self.config.base.handle_split:
            self._handle_split(trading_date)

    def after_trading(self):
        trading_date = ExecutionContext.get_current_trading_dt().date()
        portfolio = self.portfolio
        # de_listed may occur
        portfolio._portfolio_value = None

        positions = portfolio.positions

        de_listed_id_list = []
        # update buy_today_holding_quantity to zero for T+1
        for order_book_id in positions:
            position = positions[order_book_id]
            position._buy_today_holding_quantity = 0

            # 检查股票今天是否退市，如果退市，则按照close_price卖出，并warning
            if position._de_listed_date is not None and trading_date >= position._de_listed_date.date():
                de_listed_id_list.append(order_book_id)

        for de_listed_id in de_listed_id_list:
            position = positions[de_listed_id]
            if self.config.validator.cash_return_by_stock_delisted:
                portfolio._cash += position.market_value
            if position._quantity != 0:
                user_system_log.warn(
                    _("{order_book_id} is expired, close all positions by system").format(order_book_id=de_listed_id))
            del positions[de_listed_id]

    def settlement(self):
        portfolio = self.portfolio
        trading_date = ExecutionContext.get_current_trading_dt().date()

        self.portfolio_persist()

        portfolio._yesterday_portfolio_value = portfolio.portfolio_value

        self._handle_dividend_ex_dividend(trading_date)

    def bar(self, bar_dict):
        portfolio = self.portfolio
        # invalidate cache
        portfolio._portfolio_value = None
        positions = portfolio.positions

        for order_book_id, position in six.iteritems(positions):
            bar = bar_dict[order_book_id]
            if not bar.isnan:
                position._market_value = position._quantity * bar.close
                position._last_price = bar.close

    def tick(self, tick):
        portfolio = self.portfolio
        # invalidate cache
        portfolio._portfolio_value = None
        position = portfolio.positions[tick.order_book_id]

        position._market_value = position._quantity * tick.last_price
        position._last_price = tick.last_price

    def order_pending_new(self, account, order):
        if self != account:
            return
        if order._is_final():
            return
        order_book_id = order.order_book_id
        position = self.portfolio.positions[order_book_id]
        position._total_orders += 1
        create_quantity = order.quantity
        create_value = order._frozen_price * create_quantity

        self._update_order_data(order, create_quantity, create_value)
        self._update_frozen_cash(order, create_value)

    def order_creation_pass(self, account, order):
        pass

    def order_creation_reject(self, account, order):
        if self != account:
            return
        order_book_id = order.order_book_id
        position = self.portfolio.positions[order_book_id]
        position._total_orders -= 1
        cancel_quantity = order.unfilled_quantity
        cancel_value = order._frozen_price * cancel_quantity
        self._update_order_data(order, cancel_quantity, cancel_value)
        self._update_frozen_cash(order, -cancel_value)

    def order_pending_cancel(self, account, order):
        pass

    def order_cancellation_pass(self, account, order):
        self._cancel_order_cal(account, order)

    def order_cancellation_reject(self, account, order):
        pass

    def order_unsolicited_update(self, account, order):
        self._cancel_order_cal(account, order)

    def _cancel_order_cal(self, account, order):
        if self != account:
            return
        rejected_quantity = order.unfilled_quantity
        rejected_value = order._frozen_price * rejected_quantity
        self._update_order_data(order, -rejected_quantity, -rejected_value)
        self._update_frozen_cash(order, -rejected_value)

    def trade(self, account, trade):
        if self != account:
            return
        portfolio = self.portfolio
        portfolio._portfolio_value = None
        order = trade.order
        bar_dict = ExecutionContext.get_current_bar_dict()
        order_book_id = order.order_book_id
        position = portfolio.positions[order.order_book_id]
        position._is_traded = True
        trade_quantity = trade.last_quantity
        minus_value_by_trade = order._frozen_price * trade_quantity
        trade_value = trade.last_price * trade_quantity

        if order.side == SIDE.BUY:
            position._avg_price = (position._avg_price * position._quantity +
                                   trade_quantity * trade.last_price) / (position._quantity + trade_quantity)

        self._update_order_data(order, -trade_quantity, -minus_value_by_trade)
        self._update_trade_data(order, trade, trade_quantity, trade_value)
        self._update_frozen_cash(order, -minus_value_by_trade)
        price = bar_dict[order_book_id].close
        if order.side == SIDE.BUY and order.order_book_id not in \
                {'510900.XSHG', '513030.XSHG', '513100.XSHG', '513500.XSHG'}:
            position._buy_today_holding_quantity += trade_quantity
        position._market_value = (position._buy_trade_quantity - position._sell_trade_quantity) * price
        position._last_price = price
        position._total_trades += 1

        portfolio._total_tax += trade.tax
        portfolio._total_commission += trade.commission
        portfolio._cash = portfolio._cash - trade.tax - trade.commission
        if order.side == SIDE.BUY:
            portfolio._cash -= trade_value
        else:
            portfolio._cash += trade_value

        self._last_trade_id = trade.exec_id

    def _update_order_data(self, order, inc_order_quantity, inc_order_value):
        position = self.portfolio.positions[order.order_book_id]
        if order.side == SIDE.BUY:
            position._buy_order_quantity += inc_order_quantity
            position._buy_order_value += inc_order_value
        else:
            position._sell_order_quantity += inc_order_quantity
            position._sell_order_value += inc_order_value

    def _update_trade_data(self, order, trade, trade_quantity, trade_value):
        position = self.portfolio.positions[order.order_book_id]
        position._transaction_cost = position._transaction_cost + trade.commission + trade.tax
        if order.side == SIDE.BUY:
            position._buy_trade_quantity += trade_quantity
            position._buy_trade_value += trade_value
        else:
            position._sell_trade_quantity += trade_quantity
            position._sell_trade_value += trade_value

    def _update_frozen_cash(self, order, inc_order_value):
        portfolio = self.portfolio
        if order.side == SIDE.BUY:
            portfolio._frozen_cash += inc_order_value
            portfolio._cash -= inc_order_value

    def _handle_split(self, trading_date):
        import rqdatac
        for order_book_id, position in six.iteritems(self.portfolio.positions):
            split_df = rqdatac.get_split(order_book_id, start_date="2005-01-01", end_date="2099-01-01")
            if split_df is None:
                system_log.warn(_("no split data {}").foramt(order_book_id))
                continue
            try:
                series = split_df.loc[trading_date]
            except KeyError:
                continue

            # 处理拆股

            user_system_log.info(_("split {order_book_id}, {position}").format(
                order_book_id=order_book_id,
                position=position,
            ))

            ratio = series.split_coefficient_to / series.split_coefficient_from
            for key in ["_buy_order_quantity", "_sell_order_quantity", "_buy_trade_quantity", "_sell_trade_quantity"]:
                setattr(position, key, getattr(position, key) * ratio)

            user_system_log.info(_("split {order_book_id}, {position}").format(
                order_book_id=order_book_id,
                position=position,
            ))

            user_system_log.info(_("split {order_book_id}, {series}").format(
                order_book_id=order_book_id,
                series=series,
            ))

    def _handle_dividend_payable(self, trading_date):
        """handle dividend payable before trading
        """
        to_delete_dividend = []
        for order_book_id, dividend_info in six.iteritems(self.portfolio._dividend_info):
            dividend_series_dict = dividend_info.dividend_series_dict

            if pd.Timestamp(trading_date) == pd.Timestamp(dividend_series_dict['payable_date']):
                dividend_per_share = dividend_series_dict["dividend_cash_before_tax"] / dividend_series_dict["round_lot"]
                if dividend_per_share > 0 and dividend_info.quantity > 0:
                    dividend_cash = dividend_per_share * dividend_info.quantity
                    self.portfolio._dividend_receivable -= dividend_cash
                    self.portfolio._cash += dividend_cash
                    to_delete_dividend.append(order_book_id)

        for order_book_id in to_delete_dividend:
            self.portfolio._dividend_info.pop(order_book_id, None)

    def _handle_dividend_ex_dividend(self, trading_date):
        data_proxy = ExecutionContext.get_data_proxy()
        for order_book_id, position in six.iteritems(self.portfolio.positions):
            dividend_series = data_proxy.get_dividend_by_book_date(order_book_id, trading_date)
            if dividend_series is None:
                continue

            dividend_series_dict = {
                'book_closure_date': dividend_series['book_closure_date'].to_pydatetime(),
                'ex_dividend_date': dividend_series['ex_dividend_date'].to_pydatetime(),
                'payable_date': dividend_series['payable_date'].to_pydatetime(),
                'dividend_cash_before_tax': float(dividend_series['dividend_cash_before_tax']),
                'round_lot': int(dividend_series['round_lot'])
            }

            dividend_per_share = dividend_series_dict["dividend_cash_before_tax"] / dividend_series_dict["round_lot"]

            self.portfolio._dividend_info[order_book_id] = Dividend(order_book_id, position._quantity, dividend_series_dict)
            self.portfolio._dividend_receivable += dividend_per_share * position._quantity



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils.repr import property_repr, dict_repr


class PositionClone(object):
    __repr__ = dict_repr

    def __simple_object__(self):
        return self.__dict__


class BasePosition(object):

    __repr__ = property_repr

    def __init__(self, order_book_id):
        self._order_book_id = order_book_id
        self._last_price = 0
        self._market_value = 0.
        self._buy_trade_value = 0.
        self._sell_trade_value = 0
        self._buy_order_value = 0.
        self._sell_order_value = 0.

        self._buy_order_quantity = 0
        self._sell_order_quantity = 0
        self._buy_trade_quantity = 0
        self._sell_trade_quantity = 0

        self._total_orders = 0
        self._total_trades = 0

        self._is_traded = False

    @property
    def market_value(self):
        """
        【float】投资组合当前所有证券仓位的市值的加总
        """
        return self._market_value

    @property
    def order_book_id(self):
        """
        【str】合约代码
        """
        return self._order_book_id

    @property
    def total_orders(self):
        """
        【int】该仓位的总订单的次数
        """
        return self._total_orders

    @property
    def total_trades(self):
        """
        【int】该仓位的总成交的次数
        """
        return self._total_trades

    @property
    def _position_value(self):
        raise NotImplementedError

    @property
    def pnl(self):
        """
        【float】持仓累计盈亏
        """
        return self._market_value + self._sell_trade_value - self._buy_trade_value

    def _clone(self):
        p = PositionClone()
        for key in dir(self):
            if "__" in key:
                continue
            setattr(p, key, getattr(self, key))
        return p




# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jsonpickle

from rqalpha.interface import AbstractBroker, Persistable
from rqalpha.utils import get_account_type
from rqalpha.utils.i18n import gettext as _
from rqalpha.events import EVENT
from rqalpha.const import MATCHING_TYPE, ORDER_STATUS
from rqalpha.const import ACCOUNT_TYPE
from rqalpha.environment import Environment
from rqalpha.model.account import BenchmarkAccount, StockAccount, FutureAccount

from .matcher import Matcher


def init_accounts(env):
    accounts = {}
    config = env.config
    start_date = config.base.start_date
    total_cash = 0
    for account_type in config.base.account_list:
        if account_type == ACCOUNT_TYPE.STOCK:
            stock_starting_cash = config.base.stock_starting_cash
            accounts[ACCOUNT_TYPE.STOCK] = StockAccount(env, stock_starting_cash, start_date)
            total_cash += stock_starting_cash
        elif account_type == ACCOUNT_TYPE.FUTURE:
            future_starting_cash = config.base.future_starting_cash
            accounts[ACCOUNT_TYPE.FUTURE] = FutureAccount(env, future_starting_cash, start_date)
            total_cash += future_starting_cash
        else:
            raise NotImplementedError
    if config.base.benchmark is not None:
        accounts[ACCOUNT_TYPE.BENCHMARK] = BenchmarkAccount(env, total_cash, start_date)

    return accounts


class SimulationBroker(AbstractBroker, Persistable):
    def __init__(self, env):
        self._env = env
        if env.config.base.matching_type == MATCHING_TYPE.CURRENT_BAR_CLOSE:
            self._matcher = Matcher(lambda bar: bar.close, env.config.validator.bar_limit)
            self._match_immediately = True
        else:
            self._matcher = Matcher(lambda bar: bar.open, env.config.validator.bar_limit)
            self._match_immediately = False

        self._accounts = None
        self._open_orders = []
        self._board = None
        self._turnover = {}
        self._delayed_orders = []
        self._frontend_validator = {}

        # 该事件会触发策略的before_trading函数
        self._env.event_bus.add_listener(EVENT.BEFORE_TRADING, self.before_trading)
        # 该事件会触发策略的handle_bar函数
        self._env.event_bus.add_listener(EVENT.BAR, self.bar)
        # 该事件会触发策略的handel_tick函数
        self._env.event_bus.add_listener(EVENT.TICK, self.tick)
        # 该事件会触发策略的after_trading函数
        self._env.event_bus.add_listener(EVENT.AFTER_TRADING, self.after_trading)

    def get_accounts(self):
        if self._accounts is None:
            self._accounts = init_accounts(self._env)
        return self._accounts

    def get_open_orders(self):
        return self._open_orders

    def get_state(self):
        return jsonpickle.dumps([o.order_id for _, o in self._delayed_orders]).encode('utf-8')

    def set_state(self, state):
        delayed_orders = jsonpickle.loads(state.decode('utf-8'))
        for account in self._accounts.values():
            for o in account.daily_orders.values():
                if not o._is_final():
                    if o.order_id in delayed_orders:
                        self._delayed_orders.append((account, o))
                    else:
                        self._open_orders.append((account, o))

    def _get_account_for(self, order_book_id):
        account_type = get_account_type(order_book_id)
        return self._accounts[account_type]

    def submit_order(self, order):
        account = self._get_account_for(order.order_book_id)

        self._env.event_bus.publish_event(EVENT.ORDER_PENDING_NEW, account, order)

        account.append_order(order)
        if order._is_final():
            return

        # account.on_order_creating(order)
        if self._env.config.base.frequency == '1d' and not self._match_immediately:
            self._delayed_orders.append((account, order))
            return

        self._open_orders.append((account, order))
        order._active()
        self._env.event_bus.publish_event(EVENT.ORDER_CREATION_PASS, account, order)
        if self._match_immediately:
            self._match()

    def cancel_order(self, order):
        account = self._get_account_for(order.order_book_id)

        self._env.event_bus.publish_event(EVENT.ORDER_PENDING_CANCEL, account, order)

        # account.on_order_cancelling(order)
        order._mark_cancelled(_("{order_id} order has been cancelled by user.").format(order_id=order.order_id))

        self._env.event_bus.publish_event(EVENT.ORDER_CANCELLATION_PASS, account, order)

        # account.on_order_cancellation_pass(order)
        try:
            self._open_orders.remove((account, order))
        except ValueError:
            try:
                self._delayed_orders.remove((account, order))
            except ValueError:
                pass

    def before_trading(self):
        for account, order in self._open_orders:
            order._active()
            self._env.event_bus.publish_event(EVENT.ORDER_CREATION_PASS, account, order)

    def after_trading(self):
        for account, order in self._open_orders:
            order._mark_rejected(_("Order Rejected: {order_book_id} can not match. Market close.").format(
                order_book_id=order.order_book_id
            ))
            self._env.event_bus.publish_event(EVENT.ORDER_UNSOLICITED_UPDATE, account, order)
        self._open_orders = self._delayed_orders
        self._delayed_orders = []

    def bar(self, bar_dict):
        env = Environment.get_instance()
        self._matcher.update(env.calendar_dt, env.trading_dt, bar_dict)
        self._match()

    def tick(self, tick):
        # TODO support tick matching
        pass
        # env = Environment.get_instance()
        # self._matcher.update(env.calendar_dt, env.trading_dt, tick)
        # self._match()

    def _match(self):
        self._matcher.match(self._open_orders)
        final_orders = [(a, o) for a, o in self._open_orders if o._is_final()]
        self._open_orders = [(a, o) for a, o in self._open_orders if not o._is_final()]

        for account, order in final_orders:
            if order.status == ORDER_STATUS.REJECTED or order.status == ORDER_STATUS.CANCELLED:
                self._env.event_bus.publish_event(EVENT.ORDER_UNSOLICITED_UPDATE, account, order)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import pickle

from ..const import ACCOUNT_TYPE
from ..environment import Environment
from ..utils.logger import user_system_log, system_log
from ..utils.i18n import gettext as _
from ..utils.repr import property_repr
from ..utils.proxy import PortfolioProxy


class RunInfo(object):
    """
    策略运行信息
    """
    __repr__ = property_repr

    def __init__(self, config):
        self._run_id = config.base.run_id
        self._start_date = config.base.start_date
        self._end_date = config.base.end_date
        self._frequency = config.base.frequency
        self._stock_starting_cash = config.base.stock_starting_cash
        self._future_starting_cash = config.base.future_starting_cash
        self._slippage = config.base.slippage
        self._benchmark = config.base.benchmark
        self._matching_type = config.base.matching_type
        self._commission_multiplier = config.base.commission_multiplier
        self._margin_multiplier = config.base.margin_multiplier
        self._run_type = config.base.run_type

    @property
    def run_id(self):
        """
        :property getter: 标识策略每次运行的唯一id
        """
        return self._run_id

    @property
    def start_date(self):
        """
        :property getter: 策略的开始日期
        """
        return self._start_date

    @property
    def end_date(self):
        """
        :property getter: 策略的结束日期
        """
        return self._end_date

    @property
    def frequency(self):
        """
        :property getter: 策略频率，'1d'或'1m'
        """
        return self._frequency

    @property
    def stock_starting_cash(self):
        """
        :property getter: 股票账户初始资金
        """
        return self._stock_starting_cash

    @property
    def future_starting_cash(self):
        """
        :property getter: 期货账户初始资金
        """
        return self._future_starting_cash

    @property
    def slippage(self):
        """
        :property getter: 滑点水平
        """
        return self._slippage

    @property
    def benchmark(self):
        """
        :property getter: 基准合约代码
        """
        return self._benchmark

    @property
    def matching_type(self):
        """
        :property getter: 撮合方式
        """
        return self._matching_type

    @property
    def commission_multiplier(self):
        """
        :property getter: 手续费倍率
        """
        return self._commission_multiplier

    @property
    def margin_multiplier(self):
        """
        :property getter: 保证金倍率
        """
        return self._margin_multiplier

    @property
    def run_type(self):
        """
        :property getter: 运行类型
        """
        return self._run_type


class StrategyContext(object):
    def __repr__(self):
        items = ("%s = %r" % (k, v)
                 for k, v in six.iteritems(self.__dict__)
                 if not callable(v) and not k.startswith("_"))
        return "Context({%s})" % (', '.join(items),)

    def __init__(self):
        self._proxy_portfolio_dict = None
        self._config = None

    def get_state(self):
        dict_data = {}
        for key, value in six.iteritems(self.__dict__):
            if key.startswith("_"):
                continue
            try:
                dict_data[key] = pickle.dumps(value)
            except Exception as e:
                user_system_log.warn("context.{} can not pickle", key)
        return pickle.dumps(dict_data)

    def set_state(self, state):
        dict_data = pickle.loads(state)
        for key, value in six.iteritems(dict_data):
            try:
                self.__dict__[key] = pickle.loads(value)
                system_log.debug("restore context.{} {}", key, type(self.__dict__[key]))
            except Exception as e:
                user_system_log.warn('context.{} can not restore', key)

    @property
    def universe(self):
        """
        在运行 :func:`update_universe`, :func:`subscribe` 或者 :func:`unsubscribe` 的时候，合约池会被更新。

        需要注意，合约池内合约的交易时间（包含股票的策略默认会在股票交易时段触发）是handle_bar被触发的依据。

        :property getter: list[`str`]
        """
        return Environment.get_instance().universe

    @property
    def now(self):
        """
        使用以上的方式就可以在handle_bar中拿到当前的bar的时间，比如day bar的话就是那天的时间，minute bar的话就是这一分钟的时间点。
        """
        return Environment.get_instance().calendar_dt

    @property
    def run_info(self):
        """
        :property getter: :class:`~RunInfo`
        """
        config = Environment.get_instance().config
        return RunInfo(config)

    @property
    def portfolio(self):
        """
        该投资组合在单一股票或期货策略中分别为股票投资组合和期货投资组合。在股票+期货的混合策略中代表汇总之后的总投资组合。

        =========================   =========================   ==============================================================================
        属性                         类型                        注释
        =========================   =========================   ==============================================================================
        starting_cash               float                       初始资金，为子组合初始资金的加总
        cash                        float                       可用资金，为子组合可用资金的加总
        total_returns               float                       投资组合至今的累积收益率。计算方法是现在的投资组合价值/投资组合的初始资金
        daily_returns               float                       投资组合每日收益率
        daily_pnl                   float                       当日盈亏，子组合当日盈亏的加总
        market_value                float                       投资组合当前的市场价值，为子组合市场价值的加总
        portfolio_value             float                       总权益，为子组合总权益加总
        pnl                         float                       当前投资组合的累计盈亏
        start_date                  datetime.datetime           策略投资组合的回测/实时模拟交易的开始日期
        annualized_returns          float                       投资组合的年化收益率
        positions                   dict                        一个包含所有仓位的字典，以order_book_id作为键，position对象作为值
        =========================   =========================   ==============================================================================

        :property getter: :class:`~MixedPortfolio` | :class:`~StockPortfolio` | :class:`~FuturePortfolio`
        """
        env = Environment.get_instance()
        account_list = env.config.base.account_list
        if len(account_list) == 1:
            if account_list[0] == ACCOUNT_TYPE.STOCK:
                return self.stock_portfolio
            elif account_list[0] == ACCOUNT_TYPE.FUTURE:
                return self.future_portfolio
        return Environment.get_instance().account.portfolio

    @property
    def stock_portfolio(self):
        """
        获取股票投资组合信息。

        在单独股票类型策略中，内容与portfolio一致，都代表当前投资组合；在期货+股票混合策略中代表股票子组合；在单独期货策略中，不能被访问。

        =========================   =========================   ==============================================================================
        属性                         类型                        注释
        =========================   =========================   ==============================================================================
        starting_cash               float                       回测或实盘交易给算法策略设置的初始资金
        cash                        float                       可用资金
        total_returns               float                       投资组合至今的累积收益率。计算方法是现在的投资组合价值/投资组合的初始资金
        daily_returns               float                       当前最新一天的每日收益
        daily_pnl                   float                       当日盈亏，当日投资组合总权益-昨日投资组合总权益
        market_value                float                       投资组合当前所有证券仓位的市值的加总
        portfolio_value             float                       总权益，包含市场价值和剩余现金
        pnl                         float                       当前投资组合的累计盈亏
        start_date                  date                        策略投资组合的回测/实时模拟交易的开始日期
        annualized_returns          float                       投资组合的年化收益率
        positions                   dict                        一个包含所有证券仓位的字典，以order_book_id作为键，position对象作为值
        dividend_receivable         float                       投资组合在分红现金收到账面之前的应收分红部分
        =========================   =========================   ==============================================================================

        :property getter: :class:`~StockPortfolio`
        """
        if getattr(self, "_proxy_portfolio_dict", None) is None:
            self._proxy_portfolio_dict = {}
        self._proxy_portfolio_dict[ACCOUNT_TYPE.STOCK] = PortfolioProxy(
            Environment.get_instance().accounts[ACCOUNT_TYPE.STOCK].portfolio)
        return self._proxy_portfolio_dict[ACCOUNT_TYPE.STOCK]

    @property
    def future_portfolio(self):
        """
        获取期货投资组合信息。

        在单独期货类型策略中，内容与portfolio一致，都代表当前投资组合；在期货+股票混合策略中代表期货子组合；在单独股票策略中，不能被访问。

        =========================   =========================   ==============================================================================
        属性                         类型                        注释
        =========================   =========================   ==============================================================================
        starting_cash               float                       初始资金
        cash                        float                       可用资金
        frozen_cash                 float                       冻结资金
        total_returns               float                       投资组合至今的累积收益率，当前总权益/初始资金
        daily_returns               float                       当日收益率 = 当日收益 / 昨日总权益
        market_value                float                       投资组合当前所有期货仓位的名义市值的加总
        pnl                         float                       累计盈亏，当前投资组合总权益-初始资金
        daily_pnl                   float                       当日盈亏，当日浮动盈亏 + 当日平仓盈亏 - 当日费用
        daily_holding_pnl           float                       当日浮动盈亏
        daily_realized_pnl          float                       当日平仓盈亏
        portfolio_value             float                       总权益，昨日总权益+当日盈亏
        transaction_cost            float                       总费用
        start_date                  date                        回测开始日期
        annualized_returns          float                       投资组合的年化收益率
        positions                   dict                        一个包含期货仓位的字典，以order_book_id作为键，position对象作为值
        margin                      float                       已占用保证金
        =========================   =========================   ==============================================================================

        :property getter: :class:`~FuturePortfolio`
        """
        if getattr(self, "_proxy_portfolio_dict", None) is None:
            self._proxy_portfolio_dict = {}
        self._proxy_portfolio_dict[ACCOUNT_TYPE.FUTURE] = PortfolioProxy(
            Environment.get_instance().accounts[ACCOUNT_TYPE.FUTURE].portfolio)
        return self._proxy_portfolio_dict[ACCOUNT_TYPE.FUTURE]

    @property
    def slippage(self):
        raise NotImplementedError

    @property
    def benchmark(self):
        raise NotImplementedError

    @property
    def margin_rate(self):
        raise NotImplementedError

    @property
    def commission(self):
        raise NotImplementedError

    @property
    def short_selling_allowed(self):
        raise NotImplementedError

    @slippage.setter
    def slippage(self, value):
        user_system_log.warn(_("[deprecated] {} is no longer used.").format('context.slippage'))

    @benchmark.setter
    def benchmark(self, value):
        user_system_log.warn(_("[deprecated] {} is no longer used.").format('context.benchmark'))

    @margin_rate.setter
    def margin_rate(self, value):
        user_system_log.warn(_("[deprecated] {} is no longer used.").format('context.margin_rate'))

    @commission.setter
    def commission(self, value):
        user_system_log.warn(_("[deprecated] {} is no longer used.").format('context.commission'))

    @short_selling_allowed.setter
    def short_selling_allowed(self, value):
        user_system_log.warn(_("[deprecated] {} is no longer used.").format('context.short_selling_allowed'))



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class CustomEnum(Enum):
    def __repr__(self):
        return "%s.%s" % (
            self.__class__.__name__, self._name_)


class EXECUTION_PHASE(CustomEnum):
    GLOBAL = "[全局]"
    ON_INIT = "[程序初始化]"
    BEFORE_TRADING = "[日内交易前]"
    ON_BAR = "[盘中 handle_bar 函数]"
    ON_TICK = "[盘中 handle_tick 函数]"
    AFTER_TRADING = "[日内交易后]"
    FINALIZED = "[程序结束]"
    SCHEDULED = "[scheduler函数内]"

RUN_TYPE = CustomEnum("RUN_TYPE", [
    # Back Test
    "BACKTEST",
    # Paper Trading
    "PAPER_TRADING",
])

ACCOUNT_TYPE = CustomEnum("ACCOUNT_TYPE", [
    "TOTAL",
    "BENCHMARK",
    "STOCK",
    "FUTURE",
])

BAR_STATUS = CustomEnum("BAR_STATUS", [
    "LIMIT_UP",
    "LIMIT_DOWN",
    "NORMAL",
    "ERROR",
])

MATCHING_TYPE = CustomEnum("MATCHING_TYPE", [
    "CURRENT_BAR_CLOSE",
    "NEXT_BAR_OPEN",
])

ORDER_TYPE = CustomEnum("ORDER_TYPE", [
    "MARKET",
    "LIMIT",
])

ORDER_STATUS = CustomEnum("ORDER_STATUS", [
    "PENDING_NEW",
    "ACTIVE",
    "FILLED",
    "REJECTED",
    "PENDING_CANCEL",
    "CANCELLED",
])

SIDE = CustomEnum("SIDE", [
    "BUY",
    "SELL",
])

POSITION_EFFECT = CustomEnum("POSITION_EFFECT", [
    "OPEN",
    "CLOSE",
    "CLOSE_TODAY",
])

EXC_TYPE = CustomEnum("EXC_TYPE", [
    "USER_EXC",
    "SYSTEM_EXC",
    "NOTSET",
])

INSTRUMENT_TYPE = CustomEnum("INSTRUMENT_TYPE", [
    "CS",
    "FUTURE",
    "OPTION",
    "ETF",
    "LOF",
    "INDX",
    "FENJI_MU",
    "FENJI_A",
    "FENJI_B",
])

PERSIST_MODE = CustomEnum("PERSIST_MODE", [
    "ON_CRASH",
    "REAL_TIME"
])

MARGIN_TYPE = CustomEnum("MARGIN_TYPE", [
    "BY_MONEY",
    "BY_VOLUME",
])

COMMISSION_TYPE = CustomEnum("COMMISSION_TYPE", [
    "BY_MONEY",
    "BY_VOLUME",
])

EXIT_CODE = CustomEnum("EXIT_CODE", [
    "EXIT_SUCCESS",
    "EXIT_USER_ERROR",
    "EXIT_INTERNAL_ERROR",
])


class HEDGE_TYPE(CustomEnum):
    HEDGE = "hedge"
    SPECULATION = "speculation"
    ARBITRAGE = "arbitrage"


class DAYS_CNT(object):
    DAYS_A_YEAR = 365
    TRADING_DAYS_A_YEAR = 252


UNDERLYING_SYMBOL_PATTERN = "([a-zA-Z]+)\d+"

NIGHT_TRADING_NS = ["CU", "AL", "ZN", "PB", "SN", "NI", "RB", "HC", "BU", "RU", "AU", "AG", "Y", "M", "A", "B", "P",
                    "J", "JM", "I", "CF", "SR", "OI", "MA", "ZC", "FG", "RM"]



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
更多描述请见
https://www.ricequant.com/api/python/chn
'''

from decimal import Decimal, getcontext

import six

from .api_base import decorate_api_exc, instruments
from ..const import ACCOUNT_TYPE
from ..const import EXECUTION_PHASE, SIDE, ORDER_TYPE
from ..environment import Environment
from ..execution_context import ExecutionContext
from ..model.instrument import Instrument
from ..model.order import Order, OrderStyle, MarketOrder, LimitOrder
from ..utils.arg_checker import apply_rules, verify_that
from ..utils.exception import patch_user_exc, RQInvalidArgument
from ..utils.i18n import gettext as _
from ..utils.logger import user_system_log
from ..utils.scheduler import market_close, market_open
from ..utils import scheduler

# 使用Decimal 解决浮点数运算精度问题
getcontext().prec = 10

__all__ = [
    'market_open',
    'market_close',
    'scheduler',
]


def export_as_api(func):
    __all__.append(func.__name__)

    func = decorate_api_exc(func)

    return func


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_ins').is_valid_stock(),
             verify_that('amount').is_number(),
             verify_that('style').is_instance_of((MarketOrder, LimitOrder)))
def order_shares(id_or_ins, amount, style=MarketOrder()):
    """
    落指定股数的买/卖单，最常见的落单方式之一。如有需要落单类型当做一个参量传入，如果忽略掉落单类型，那么默认是市价单（market order）。

    :param id_or_ins: 下单标的物
    :type id_or_ins: :class:`~Instrument` object | `str`

    :param int amount: 下单量, 正数代表买入，负数代表卖出。将会根据一手xx股来向下调整到一手的倍数，比如中国A股就是调整成100股的倍数。

    :param style: 下单类型, 默认是市价单。目前支持的订单类型有 :class:`~LimitOrder` 和 :class:`~MarketOrder`
    :type style: `OrderStyle` object

    :return: :class:`~Order` object

    :example:

    .. code-block:: python

        #购买Buy 2000 股的平安银行股票，并以市价单发送：
        order_shares('000001.XSHE', 2000)
        #卖出2000股的平安银行股票，并以市价单发送：
        order_shares('000001.XSHE', -2000)
        #购买1000股的平安银行股票，并以限价单发送，价格为￥10：
        order_shares('000001.XSHG', 1000, style=LimitOrder(10))
    """
    # Place an order by specified number of shares. Order type is also
    #     passed in as parameters if needed. If style is omitted, it fires a
    #     market order by default.
    # :PARAM id_or_ins: the instrument to be ordered
    # :type id_or_ins: str or Instrument
    # :param float amount: Number of shares to order. Positive means buy,
    #     negative means sell. It will be rounded down to the closest
    #     integral multiple of the lot size
    # :param style: Order type and default is `MarketOrder()`. The
    #     available order types are: `MarketOrder()` and
    #     `LimitOrder(limit_price)`
    # :return:  A unique order id.
    # :rtype: int
    if not isinstance(style, OrderStyle):
        raise RQInvalidArgument(_('style should be OrderStyle'))
    if isinstance(style, LimitOrder):
        if style.get_limit_price() <= 0:
            raise RQInvalidArgument(_("Limit order price should be positive"))

    order_book_id = assure_stock_order_book_id(id_or_ins)
    bar_dict = ExecutionContext.get_current_bar_dict()
    bar = bar_dict[order_book_id]
    price = bar.close
    calendar_dt = ExecutionContext.get_current_calendar_dt()
    trading_dt = ExecutionContext.get_current_trading_dt()

    if amount > 0:
        side = SIDE.BUY
    else:
        amount = abs(amount)
        side = SIDE.SELL

    round_lot = int(ExecutionContext.data_proxy.instruments(order_book_id).round_lot)

    try:
        amount = int(Decimal(amount) / Decimal(round_lot)) * round_lot
    except ValueError:
        amount = 0

    r_order = Order.__from_create__(calendar_dt, trading_dt, order_book_id, amount, side, style, None)

    if bar.isnan or price == 0:
        user_system_log.warn(_("Order Creation Failed: [{order_book_id}] No market data").format(order_book_id=order_book_id))
        r_order._mark_rejected(_("Order Creation Failed: [{order_book_id}] No market data").format(order_book_id=order_book_id))
        return r_order

    if amount == 0:
        # 如果计算出来的下单量为0, 则不生成Order, 直接返回None
        # 因为很多策略会直接在handle_bar里面执行order_target_percent之类的函数，经常会出现下一个量为0的订单，如果这些订单都生成是没有意义的。
        r_order._mark_rejected(_("Order Creation Failed: 0 order quantity"))
        return r_order
    if r_order.type == ORDER_TYPE.MARKET:
        bar_dict = ExecutionContext.get_current_bar_dict()
        r_order._frozen_price = bar_dict[order_book_id].close
    ExecutionContext.broker.submit_order(r_order)

    return r_order


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_ins').is_valid_stock(),
             verify_that('amount').is_number(),
             verify_that('style').is_instance_of((MarketOrder, LimitOrder)))
def order_lots(id_or_ins, amount, style=MarketOrder()):
    """
    指定手数发送买/卖单。如有需要落单类型当做一个参量传入，如果忽略掉落单类型，那么默认是市价单（market order）。

    :param id_or_ins: 下单标的物
    :type id_or_ins: :class:`~Instrument` object | `str`

    :param int amount: 下单量, 正数代表买入，负数代表卖出。将会根据一手xx股来向下调整到一手的倍数，比如中国A股就是调整成100股的倍数。

    :param style: 下单类型, 默认是市价单。目前支持的订单类型有 :class:`~LimitOrder` 和 :class:`~MarketOrder`
    :type style: `OrderStyle` object

    :return: :class:`~Order` object

    :example:

    .. code-block:: python

        #买入20手的平安银行股票，并且发送市价单：
        order_lots('000001.XSHE', 20)
        #买入10手平安银行股票，并且发送限价单，价格为￥10：
        order_lots('000001.XSHE', 10, style=LimitOrder(10))

    """
    # Place an order by specified number of lots. Order type is also passed
    #     in as parameters if needed. If style is omitted, it fires a market
    #     order by default.
    # :param id_or_ins: the instrument to be ordered
    # :type id_or_ins: str or Instrument
    # :param float amount: Number of lots to order. Positive means buy,
    #     negative means sell.
    # :param style: Order type and default is `MarketOrder()`. The
    #     available order types are: `MarketOrder()` and
    #     `LimitOrder(limit_price)`
    # :return:  A unique order id.
    # :rtype: int
    order_book_id = assure_stock_order_book_id(id_or_ins)

    round_lot = int(ExecutionContext.get_instrument(order_book_id).round_lot)

    return order_shares(id_or_ins, amount * round_lot, style)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_ins').is_valid_stock(),
             verify_that('cash_amount').is_number(),
             verify_that('style').is_instance_of((MarketOrder, LimitOrder)))
def order_value(id_or_ins, cash_amount, style=MarketOrder()):
    """
    使用想要花费的金钱买入/卖出股票，而不是买入/卖出想要的股数，正数代表买入，负数代表卖出。股票的股数总是会被调整成对应的100的倍数（在A中国A股市场1手是100股）。当您提交一个卖单时，该方法代表的意义是您希望通过卖出该股票套现的金额。如果金额超出了您所持有股票的价值，那么您将卖出所有股票。需要注意，如果资金不足，该API将不会创建发送订单。

    :param id_or_ins: 下单标的物
    :type id_or_ins: :class:`~Instrument` object | `str`

    :param float cash_amount: 需要花费现金购买/卖出证券的数目。正数代表买入，负数代表卖出。

    :param style: 下单类型, 默认是市价单。目前支持的订单类型有 :class:`~LimitOrder` 和 :class:`~MarketOrder`
    :type style: `OrderStyle` object

    :return: :class:`~Order` object

    :example:

    .. code-block:: python

        #买入价值￥10000的平安银行股票，并以市价单发送。如果现在平安银行股票的价格是￥7.5，那么下面的代码会买入1300股的平安银行，因为少于100股的数目将会被自动删除掉：
        order_value('000001.XSHE', 10000)
        #卖出价值￥10000的现在持有的平安银行：
        order_value('000001.XSHE', -10000)

    """
    # Place an order by specified value amount rather than specific number
    #     of shares/lots. Negative cash_amount results in selling the given
    #     amount of value, if the cash_amount is larger than you current
    #     security’s position, then it will sell all shares of this security.
    #     Orders are always truncated to whole lot shares.
    # :param id_or_ins: the instrument to be ordered
    # :type id_or_ins: str or Instrument
    # :param float cash_amount: Cash amount to buy / sell the given value of
    #     securities. Positive means buy, negative means sell.
    # :param style: Order type and default is `MarketOrder()`. The
    #     available order types are: `MarketOrder()` and
    #     `LimitOrder(limit_price)`
    # :return:  A unique order id.
    # :rtype: int
    if not isinstance(style, OrderStyle):
        raise RQInvalidArgument(_('style should be OrderStyle'))
    if isinstance(style, LimitOrder):
        if style.get_limit_price() <= 0:
            raise RQInvalidArgument(_("Limit order price should be positive"))

    order_book_id = assure_stock_order_book_id(id_or_ins)

    bar_dict = ExecutionContext.get_current_bar_dict()
    bar = bar_dict[order_book_id]
    price = bar.close

    if bar.isnan or price == 0:
        return order_shares(order_book_id, 0, style)

    account = ExecutionContext.accounts[ACCOUNT_TYPE.STOCK]
    round_lot = int(ExecutionContext.get_instrument(order_book_id).round_lot)

    if cash_amount > 0:
        cash_amount = min(cash_amount, account.portfolio.cash)

    if isinstance(style, MarketOrder):
        amount = int(Decimal(cash_amount) / Decimal(price) / Decimal(round_lot)) * round_lot
    else:
        amount = int(Decimal(cash_amount) / Decimal(style.get_limit_price()) / Decimal(round_lot)) * round_lot

    # if the cash_amount is larger than you current security’s position,
    # then it will sell all shares of this security.

    position = account.portfolio.positions[order_book_id]
    amount = downsize_amount(amount, position)

    return order_shares(order_book_id, amount, style)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_ins').is_valid_stock(),
             verify_that('percent').is_number().is_greater_than(-1).is_less_than(1),
             verify_that('style').is_instance_of((MarketOrder, LimitOrder)))
def order_percent(id_or_ins, percent, style=MarketOrder()):
    """
    发送一个等于目前投资组合价值（市场价值和目前现金的总和）一定百分比的买/卖单，正数代表买，负数代表卖。股票的股数总是会被调整成对应的一手的股票数的倍数（1手是100股）。百分比是一个小数，并且小于或等于1（<=100%），0.5表示的是50%.需要注意，如果资金不足，该API将不会创建发送订单。

    :param id_or_ins: 下单标的物
    :type id_or_ins: :class:`~Instrument` object | `str`

    :param float percent: 占有现有的投资组合价值的百分比。正数表示买入，负数表示卖出。

    :param style: 下单类型, 默认是市价单。目前支持的订单类型有 :class:`~LimitOrder` 和 :class:`~MarketOrder`
    :type style: `OrderStyle` object

    :return: :class:`~Order` object

    :example:

    .. code-block:: python

        #买入等于现有投资组合50%价值的平安银行股票。如果现在平安银行的股价是￥10/股并且现在的投资组合总价值是￥2000，那么将会买入200股的平安银行股票。（不包含交易成本和滑点的损失）：
        order_percent('000001.XSHG', 0.5)
    """
    # Place an order for a security for a given percent of the current
    #     portfolio value, which is the sum of the positions value and
    #     ending cash balance. A negative percent order will result in
    #     selling given percent of current portfolio value. Orders are
    #     always truncated to whole shares. Percent should be a decimal
    #     number (0.50 means 50%), and its absolute value is <= 1.
    # :param id_or_ins: the instrument to be ordered
    # :type id_or_ins: str or Instrument
    # :param float percent: Percent of the current portfolio value. Positive
    #     means buy, negative means selling give percent of the current
    #     portfolio value. Orders are always truncated according to lot size.
    # :param style: Order type and default is `MarketOrder()`. The
    #     available order types are: `MarketOrder()` and
    #     `LimitOrder(limit_price)`
    # :return:  A unique order id.
    # :rtype: int
    if percent < -1 or percent > 1:
        raise RQInvalidArgument(_('percent should between -1 and 1'))

    account = ExecutionContext.accounts[ACCOUNT_TYPE.STOCK]
    portfolio_value = account.portfolio.portfolio_value
    return order_value(id_or_ins, portfolio_value * percent, style)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_ins').is_valid_stock(),
             verify_that('cash_amount').is_number(),
             verify_that('style').is_instance_of((MarketOrder, LimitOrder)))
def order_target_value(id_or_ins, cash_amount, style=MarketOrder()):
    """
    买入/卖出并且自动调整该证券的仓位到一个目标价值。如果还没有任何该证券的仓位，那么会买入全部目标价值的证券。如果已经有了该证券的仓位，则会买入/卖出调整该证券的现在仓位和目标仓位的价值差值的数目的证券。需要注意，如果资金不足，该API将不会创建发送订单。

    :param id_or_ins: 下单标的物
    :type id_or_ins: :class:`~Instrument` object | `str` | List[:class:`~Instrument`] | List[`str`]

    :param float cash_amount: 最终的该证券的仓位目标价值。

    :param style: 下单类型, 默认是市价单。目前支持的订单类型有 :class:`~LimitOrder` 和 :class:`~MarketOrder`
    :type style: `OrderStyle` object

    :return: :class:`~Order` object

    :example:

    .. code-block:: python

        #如果现在的投资组合中持有价值￥3000的平安银行股票的仓位并且设置其目标价值为￥10000，以下代码范例会发送价值￥7000的平安银行的买单到市场。（向下调整到最接近每手股数即100的倍数的股数）：
        order_target_value('000001.XSHE', 10000)
    """
    # Place an order to adjust a position to a target value. If there is no
    #     position for the security, an order is placed for the whole amount
    #     of target value. If there is already a position for the security,
    #     an order is placed for the difference between target value and
    #     current position value.
    # :param id_or_ins: the instrument to be ordered
    # :type id_or_ins: str or Instrument
    # :param float cash_amount: Target cash value for the adjusted position
    #     after placing order.
    # :param style: Order type and default is `MarketOrder()`. The
    #     available order types are: `MarketOrder()` and
    #     `LimitOrder(limit_price)`
    # :return:  A unique order id.
    # :rtype: int
    order_book_id = assure_stock_order_book_id(id_or_ins)

    bar_dict = ExecutionContext.get_current_bar_dict()
    bar = bar_dict[order_book_id]
    price = 0 if bar.isnan else bar.close

    position = ExecutionContext.accounts[ACCOUNT_TYPE.STOCK].portfolio.positions[order_book_id]

    current_value = position._quantity * price

    return order_value(order_book_id, cash_amount - current_value, style)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_ins').is_valid_stock(),
             verify_that('percent').is_number().is_greater_than(0).is_less_than(1),
             verify_that('style').is_instance_of((MarketOrder, LimitOrder)))
def order_target_percent(id_or_ins, percent, style=MarketOrder()):
    """
    买入/卖出证券以自动调整该证券的仓位到占有一个指定的投资组合的目标百分比。

    *   如果投资组合中没有任何该证券的仓位，那么会买入等于现在投资组合总价值的目标百分比的数目的证券。
    *   如果投资组合中已经拥有该证券的仓位，那么会买入/卖出目标百分比和现有百分比的差额数目的证券，最终调整该证券的仓位占据投资组合的比例至目标百分比。

    其实我们需要计算一个position_to_adjust (即应该调整的仓位)

    `position_to_adjust = target_position - current_position`

    投资组合价值等于所有已有仓位的价值和剩余现金的总和。买/卖单会被下舍入一手股数（A股是100的倍数）的倍数。目标百分比应该是一个小数，并且最大值应该<=1，比如0.5表示50%。

    如果position_to_adjust 计算之后是正的，那么会买入该证券，否则会卖出该证券。 需要注意，如果资金不足，该API将不会创建发送订单。

    :param id_or_ins: 下单标的物
    :type id_or_ins: :class:`~Instrument` object | `str` | List[:class:`~Instrument`] | List[`str`]

    :param float percent: 仓位最终所占投资组合总价值的目标百分比。

    :param style: 下单类型, 默认是市价单。目前支持的订单类型有 :class:`~LimitOrder` 和 :class:`~MarketOrder`
    :type style: `OrderStyle` object

    :return: :class:`~Order` object

    :example:

    .. code-block:: python

        #如果投资组合中已经有了平安银行股票的仓位，并且占据目前投资组合的10%的价值，那么以下代码会买入平安银行股票最终使其占据投资组合价值的15%：
        order_target_percent('000001.XSHE', 0.15)
    """
    # Place an order to adjust position to a target percent of the portfolio
    #     value, so that your final position value takes the percentage you
    #     defined of your whole portfolio.
    #     position_to_adjust = target_position - current_position
    #     Portfolio value is calculated as sum of positions value and ending
    #     cash balance. The order quantity will be rounded down to integral
    #     multiple of lot size. Percent should be a decimal number (0.50
    #     means 50%), and its absolute value is <= 1. If the
    #     position_to_adjust calculated is positive, then it fires buy
    #     orders, otherwise it fires sell orders.
    # :param id_or_ins: the instrument to be ordered
    # :type id_or_ins: str or Instrument
    # :param float percent: Number of percent to order. It will be rounded down
    #     to the closest integral multiple of the lot size
    # :param style: Order type and default is `MarketOrder()`. The
    #     available order types are: `MarketOrder()` and
    #     `LimitOrder(limit_price)`
    # :return:  A unique order id.
    # :rtype: int
    if percent < 0 or percent > 1:
        raise RQInvalidArgument(_('percent should between 0 and 1'))
    order_book_id = assure_stock_order_book_id(id_or_ins)

    bar_dict = ExecutionContext.get_current_bar_dict()
    bar = bar_dict[order_book_id]
    price = 0 if bar.isnan else bar.close

    portfolio = ExecutionContext.accounts[ACCOUNT_TYPE.STOCK].portfolio
    position = portfolio.positions[order_book_id]

    current_value = position._quantity * price

    return order_value(order_book_id, portfolio.portfolio_value * percent - current_value, style)


def assure_stock_order_book_id(id_or_symbols):
    if isinstance(id_or_symbols, Instrument):
        order_book_id = id_or_symbols.order_book_id
        """
        这所以使用XSHG和XSHE来判断是否可交易是因为股票类型策略支持很多种交易类型，比如CS, ETF, LOF, FenjiMU, FenjiA, FenjiB,
        INDX等，但实际其中部分由不能交易，所以不能直接按照类型区分该合约是否可以交易。而直接通过判断其后缀可以比较好的区分是否可以进行交易
        """
        if "XSHG" in order_book_id or "XSHE" in order_book_id:
            return order_book_id
        else:
            raise RQInvalidArgument(
                _("{order_book_id} is not supported in current strategy type").format(
                    order_book_id=order_book_id))
    elif isinstance(id_or_symbols, six.string_types):
        return assure_stock_order_book_id(instruments(id_or_symbols))
    else:
        raise RQInvalidArgument(_("unsupported order_book_id type"))


def downsize_amount(amount, position):
    config = Environment.get_instance().config
    if not config.validator.close_amount:
        return amount
    if amount > 0:
        return amount
    else:
        amount = abs(amount)
        if amount > position.sellable:
            return -position.sellable
        else:
            return -amount



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import time

from ..utils import id_gen
from ..utils.repr import property_repr, properties

TradePersistMap = {
    "_calendar_dt": "_calendar_dt",
    "_trading_dt": "_trading_dt",
    "_price": "_price",
    "_amount": "_amount",
    "_order_id": "_order_id",
    "_commission": "_commission",
    "_tax": "_tax",
    "_trade_id": "_trade_id",
    "_close_today_amount": "_close_today_amount",
}


class Trade(object):
    __slots__ = ["_calendar_dt", "_trading_dt", "_price", "_amount", "_order", "_commission", "_tax", "_trade_id",
                 "_close_today_amount"]

    __repr__ = property_repr

    trade_id_gen = id_gen(int(time.time()))

    def __init__(self):
        self._calendar_dt = None
        self._trading_dt = None
        self._price = None
        self._amount = None
        self._order = None
        self._commission = None
        self._tax = None
        self._trade_id = None
        self._close_today_amount = None

    @classmethod
    def __from_create__(cls, order, calendar_dt, trading_dt, price, amount, commission=0., tax=0., trade_id=None,
                        close_today_amount=0):
        trade = cls()
        trade._calendar_dt = calendar_dt
        trade._trading_dt = trading_dt
        trade._price = price
        trade._amount = amount
        trade._order = order
        trade._commission = commission
        trade._tax = tax
        trade._trade_id = trade_id if trade_id is not None else next(trade.trade_id_gen)
        trade._close_today_amount = close_today_amount
        return trade

    @classmethod
    def __from_dict__(cls, trade_dict, order):
        trade = cls()
        for persist_key, origin_key in six.iteritems(TradePersistMap):
            if persist_key == "_order_id":
                continue
            setattr(trade, origin_key, trade_dict[persist_key])
        trade._order = order
        return trade

    def __to_dict__(self):
        trade_dict = {}
        for persist_key, origin_key in six.iteritems(TradePersistMap):
            if persist_key == "_order_id":
                trade_dict["_order_id"] = self._order.order_id
            else:
                trade_dict[persist_key] = getattr(self, origin_key)
        return trade_dict

    @property
    def trading_datetime(self):
        return self._trading_dt

    @property
    def datetime(self):
        return self._calendar_dt

    @property
    def order_id(self):
        return self.order.order_id

    @property
    def last_price(self):
        return self._price

    @property
    def last_quantity(self):
        return self._amount

    @property
    def order(self):
        return self._order

    @property
    def commission(self):
        return self._commission

    @property
    def tax(self):
        return self._tax

    @property
    def transaction_cost(self):
        return self._tax + self._commission

    @property
    def position_effect(self):
        return self.order.position_effect

    @property
    def exec_id(self):
        return self._trade_id

    def __simple_object__(self):
        return properties(self)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import csv

from rqalpha.interface import AbstractMod
from rqalpha.events import EVENT


class ProgressiveOutputCSVMod(AbstractMod):

    def start_up(self, env, mod_config):
        self._env = env
        self._mod_config = mod_config

        env.event_bus.add_listener(EVENT.POST_BAR, self._output_feeds)

        output_path = mod_config.output_path

        filename = os.path.join(output_path, "portfolio.csv")
        new_file = False
        if not os.path.exists(filename):
            new_file = True
        self.csv_file = open(filename, 'a')
        fieldnames = ["datetime", "portfolio_value", "market_value", "total_returns"]
        self.csv_writer = csv.DictWriter(self.csv_file, fieldnames)
        if new_file:
            self.csv_writer.writeheader()

    def _output_feeds(self, *args, **kwargs):
        misc_account = self._env.account
        calendar_date = self._env.calendar_dt.date()
        portfolio = misc_account.portfolio

        self.csv_writer.writerow({
            "datetime": calendar_date,
            "total_returns": portfolio.total_returns,
            "portfolio_value": portfolio.portfolio_value,
            "market_value": portfolio.market_value,
        })
        self.csv_file.flush()

    def tear_down(self, code, exception=None):
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import import_module
from collections import OrderedDict

from rqalpha.utils.logger import system_log
from rqalpha.utils.i18n import gettext as _


class ModHandler(object):
    def __init__(self):
        self._env = None
        self._mod_list = []
        self._mod_dict = OrderedDict()

    def set_env(self, environment):
        self._env = environment

        config = environment.config

        for mod_name in config.mod.__dict__:
            mod_config = getattr(config.mod, mod_name)
            if not mod_config.enabled:
                continue
            self._mod_list.append((mod_name, mod_config))

        self._mod_list.sort(key=lambda item: item[1].priority)
        for mod_name, mod_config in self._mod_list:
            system_log.debug(_('loading mod {}').format(mod_name))
            mod_module = import_module(mod_config.lib)
            mod = mod_module.load_mod()
            self._mod_dict[mod_name] = mod

        environment.mod_dict = self._mod_dict

    def start_up(self):
        for mod_name, mod_config in self._mod_list:
            self._mod_dict[mod_name].start_up(self._env, mod_config)

    def tear_down(self, *args):
        for mod_name, _ in reversed(self._mod_list):
            self._mod_dict[mod_name].tear_down(*args)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from six import with_metaclass

from ..const import SIDE, ACCOUNT_TYPE
from ..execution_context import ExecutionContext


def init_tax(account_type, rate=None):
    if account_type in [ACCOUNT_TYPE.STOCK, ACCOUNT_TYPE.BENCHMARK]:
        if rate is None:
            rate = 0.001
        return StockTax(rate)
    elif account_type == ACCOUNT_TYPE.FUTURE:
        if rate is None:
            rate = 0.
        return FutureTax(rate)
    else:
        raise NotImplementedError


class BaseTax(with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def get_tax(self, trade):
        raise NotImplementedError


class StockTax(BaseTax):
    def __init__(self, rate):
        self.rate = rate

    def get_tax(self, trade):
        cost_money = trade.last_price * trade.last_quantity
        if ExecutionContext.get_instrument(trade.order.order_book_id).type == 'CS':
            return cost_money * self.rate if trade.order.side == SIDE.SELL else 0
        else:
            return 0


class FutureTax(BaseTax):
    def __init__(self, rate):
        self.rate = rate

    def get_tax(self, trade):
        return 0



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from collections import OrderedDict

from ..portfolio import init_portfolio
from ..slippage import init_slippage
from ..tax import init_tax
from ..trade import Trade
from ..order import Order
from ...execution_context import ExecutionContext
from ...interface import Persistable
from ...utils import json as json_utils
from ...events import EVENT


class BaseAccount(Persistable):
    def __init__(self, env, init_cash, start_date, account_type):
        self._account_type = account_type
        self._env = env
        self.config = env.config

        self.portfolio = init_portfolio(init_cash, start_date, account_type)
        self.slippage_decider = init_slippage(env.config.base.slippage)
        commission_initializer = env._commission_initializer
        self.commission_decider = commission_initializer(self._account_type, env.config.base.commission_multiplier)
        self.tax_decider = init_tax(self._account_type)

        self.all_portfolios = OrderedDict()
        self.daily_orders = {}
        self.daily_trades = []
        self._last_trade_id = 0

        # 该事件会触发策略的before_trading函数
        self._env.event_bus.add_listener(EVENT.BEFORE_TRADING, self.before_trading)
        # 该事件会触发策略的handle_bar函数
        self._env.event_bus.add_listener(EVENT.BAR, self.bar)
        # 该事件会触发策略的handel_tick函数
        self._env.event_bus.add_listener(EVENT.TICK, self.tick)
        # 该事件会触发策略的after_trading函数
        self._env.event_bus.add_listener(EVENT.AFTER_TRADING, self.after_trading)
        # 触发结算事件
        self._env.event_bus.add_listener(EVENT.SETTLEMENT, self.settlement)

        # 创建订单
        self._env.event_bus.add_listener(EVENT.ORDER_PENDING_NEW, self.order_pending_new)
        # 创建订单成功
        self._env.event_bus.add_listener(EVENT.ORDER_CREATION_PASS, self.order_creation_pass)
        # 创建订单失败
        self._env.event_bus.add_listener(EVENT.ORDER_CREATION_REJECT, self.order_creation_reject)
        # 创建撤单
        self._env.event_bus.add_listener(EVENT.ORDER_PENDING_CANCEL, self.order_pending_cancel)
        # 撤销订单成功
        self._env.event_bus.add_listener(EVENT.ORDER_CANCELLATION_PASS, self.order_cancellation_pass)
        # 撤销订单失败
        self._env.event_bus.add_listener(EVENT.ORDER_CANCELLATION_REJECT, self.order_cancellation_reject)
        # 订单状态更新
        self._env.event_bus.add_listener(EVENT.ORDER_UNSOLICITED_UPDATE, self.order_unsolicited_update)
        # 成交
        self._env.event_bus.add_listener(EVENT.TRADE, self.trade)

    def set_state(self, state):
        persist_dict = json_utils.convert_json_to_dict(state.decode('utf-8'))
        self.portfolio.restore_from_dict_(persist_dict['portfolio'])

        del self.daily_trades[:]
        self.daily_orders.clear()

        for order_id, order_dict in six.iteritems(persist_dict["daily_orders"]):
            self.daily_orders[order_id] = Order.__from_dict__(order_dict)

        for trade_dict in persist_dict["daily_trades"]:
            trade = Trade.__from_dict__(trade_dict, self.daily_orders[str(trade_dict["_order_id"])])
            self.daily_trades.append(trade)

        if 'last_trade_id' in persist_dict:
            self._last_trade_id = persist_dict['last_trade_id']

    def get_state(self):
        return json_utils.convert_dict_to_json(self.__to_dict__()).encode('utf-8')

    def __to_dict__(self):
        account_dict = {
            "portfolio": self.portfolio.__to_dict__(),
            'last_trade_id': self._last_trade_id,
            "daily_orders": {order_id: order.__to_dict__() for order_id, order in six.iteritems(self.daily_orders)},
            "daily_trades": [trade.__to_dict__() for trade in self.daily_trades],
        }
        return account_dict

    def portfolio_persist(self):
        trading_date = ExecutionContext.get_current_trading_dt().date()
        self.all_portfolios[trading_date] = self.portfolio._clone()

    def get_portfolio(self, trading_date):
        return self.all_portfolios[trading_date]

    def append_order(self, order, is_active=True):
        self.daily_orders[order.order_id] = order

    def append_trade(self, trade):
        self.daily_trades.append(trade)

    def get_open_orders(self):
        return {order.order_id: order for order in self.daily_orders.values() if order._is_active()}

    def get_order(self, order_id):
        return self.daily_orders.get(order_id, None)

    def before_trading(self):
        open_orders = {}
        for k, order in six.iteritems(self.daily_orders):
            if not order._is_final():
                open_orders[k] = order

        self.daily_orders = open_orders
        self.daily_trades = []

    def last_trade_id(self):
        return self._last_trade_id

    def bar(self, bar_dict):
        pass

    def tick(self, tick):
        pass

    def after_trading(self):
        pass

    def settlement(self):
        pass

    def order_pending_new(self, account, order):
        pass

    def order_creation_pass(self, account, order):
        pass

    def order_creation_reject(self, account, order):
        pass

    def order_pending_cancel(self, account, order):
        pass

    def order_cancellation_pass(self, account, order):
        pass

    def order_cancellation_reject(self, account, order):
        pass

    def order_unsolicited_update(self, account, order):
        pass

    def trade(self, account, trade):
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
更多描述请见
https://www.ricequant.com/api/python/chn
'''

from __future__ import division
import sys
import datetime
import inspect
import pandas as pd
import six
from collections import Iterable
from dateutil.parser import parse
from types import FunctionType
from functools import wraps
from typing import List

from ..environment import Environment
from ..model.instrument import Instrument, SectorCode as sector_code, IndustryCode as industry_code
from ..model.instrument import SectorCodeItem, IndustryCodeItem
from ..execution_context import ExecutionContext
from ..const import EXECUTION_PHASE, EXC_TYPE, ORDER_STATUS, SIDE, POSITION_EFFECT, ORDER_TYPE, MATCHING_TYPE, RUN_TYPE
from ..utils import to_industry_code, to_sector_name, unwrapper
from ..utils.exception import patch_user_exc, patch_system_exc, EXC_EXT_NAME, RQInvalidArgument
from ..utils.i18n import gettext as _
from ..model.snapshot import SnapshotObject
from ..model.order import Order, MarketOrder, LimitOrder
from ..model.slippage import PriceRatioSlippage
from ..utils.arg_checker import apply_rules, verify_that
from . import names

__all__ = [
    'sector_code',
    'industry_code',
    'LimitOrder',
    'MarketOrder',
    'PriceRatioSlippage',
    'ORDER_STATUS',
    'SIDE',
    'POSITION_EFFECT',
    'ORDER_TYPE',
    'RUN_TYPE',
    'MATCHING_TYPE',
]


def decorate_api_exc(func):
    f = func
    exception_checked = False
    while True:
        if getattr(f, '_rq_exception_checked', False):
            exception_checked = True
            break

        f = getattr(f, '__wrapped__', None)
        if f is None:
            break
    if not exception_checked:
        func = api_exc_patch(func)

    return func


def api_exc_patch(func):
    if isinstance(func, FunctionType):
        @wraps(func)
        def deco(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except RQInvalidArgument:
                raise
            except Exception as e:
                if isinstance(e, TypeError):
                    exc_info = sys.exc_info()
                    try:
                        ret = inspect.getcallargs(unwrapper(func), *args, **kwargs)
                    except TypeError:
                        t, v, tb = exc_info
                        raise patch_user_exc(v.with_traceback(tb))

                if getattr(e, EXC_EXT_NAME, EXC_TYPE.NOTSET) == EXC_TYPE.NOTSET:
                    patch_system_exc(e)

                raise

        return deco
    return func


def register_api(name, func):
    globals()[name] = func
    __all__.append(name)


def export_as_api(func):
    __all__.append(func.__name__)

    func = decorate_api_exc(func)

    return func


def assure_order_book_id(id_or_ins):
    if isinstance(id_or_ins, Instrument):
        order_book_id = id_or_ins.order_book_id
    elif isinstance(id_or_ins, six.string_types):
        order_book_id = instruments(id_or_ins).order_book_id
    else:
        raise RQInvalidArgument(_("unsupported order_book_id type"))

    return order_book_id


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
def get_order(order_id):
    if isinstance(order_id, Order):
        return order_id
    else:
        return ExecutionContext.account.get_order(order_id)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
def get_open_orders():
    """
    获取当日未成交订单数据

    :return: List[:class:`~Order` object]
    """
    return ExecutionContext.account.get_open_orders()


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
def cancel_order(order):
    """
    撤单

    :param order: 需要撤销的order对象
    :type order: :class:`~Order` object
    """
    order = order if isinstance(order, Order) else get_order(order)
    if order is None:
        patch_user_exc(KeyError(_("Cancel order fail: invalid order id")))
    ExecutionContext.broker.cancel_order(order)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_symbols').are_valid_instruments())
def update_universe(id_or_symbols):
    """
    该方法用于更新现在关注的证券的集合（e.g.：股票池）。PS：会在下一个bar事件触发时候产生（新的关注的股票池更新）效果。并且update_universe会是覆盖（overwrite）的操作而不是在已有的股票池的基础上进行增量添加。比如已有的股票池为['000001.XSHE', '000024.XSHE']然后调用了update_universe(['000030.XSHE'])之后，股票池就会变成000030.XSHE一个股票了，随后的数据更新也只会跟踪000030.XSHE这一个股票了。

    :param id_or_ins: 标的物
    :type id_or_ins: :class:`~Instrument` object | `str` | List[:class:`~Instrument`] | List[`str`]
    """
    if isinstance(id_or_symbols, (six.string_types, Instrument)):
        id_or_symbols = [id_or_symbols]
    order_book_ids = set(assure_order_book_id(order_book_id) for order_book_id in id_or_symbols)
    if order_book_ids != Environment.get_instance().universe:
        Environment.get_instance().update_universe(order_book_ids)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_symbols').are_valid_instruments())
def subscribe(id_or_symbols):
    """
    订阅合约行情。该操作会导致合约池内合约的增加，从而影响handle_bar中处理bar数据的数量。

    需要注意，用户在初次编写策略时候需要首先订阅合约行情，否则handle_bar不会被触发。

    :param id_or_ins: 标的物
    :type id_or_ins: :class:`~Instrument` object | `str` | List[:class:`~Instrument`] | List[`str`]
    """
    current_universe = Environment.get_instance().universe
    if isinstance(id_or_symbols, six.string_types):
        order_book_id = instruments(id_or_symbols).order_book_id
        current_universe.add(order_book_id)
    elif isinstance(id_or_symbols, Instrument):
        current_universe.add(id_or_symbols.order_book_id)
    elif isinstance(id_or_symbols, Iterable):
        for item in id_or_symbols:
            current_universe.add(assure_order_book_id(item))
    else:
        raise RQInvalidArgument(_("unsupported order_book_id type"))
    verify_that('id_or_symbols')._are_valid_instruments("subscribe", id_or_symbols)
    Environment.get_instance().update_universe(current_universe)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_symbols').are_valid_instruments())
def unsubscribe(id_or_symbols):
    """
    取消订阅合约行情。取消订阅会导致合约池内合约的减少，如果当前合约池中没有任何合约，则策略直接退出。

    :param id_or_ins: 标的物
    :type id_or_ins: :class:`~Instrument` object | `str` | List[:class:`~Instrument`] | List[`str`]
    """
    current_universe = Environment.get_instance().universe
    if isinstance(id_or_symbols, six.string_types):
        order_book_id = instruments(id_or_symbols).order_book_id
        current_universe.discard(order_book_id)
    elif isinstance(id_or_symbols, Instrument):
        current_universe.discard(id_or_symbols.order_book_id)
    elif isinstance(id_or_symbols, Iterable):
        for item in id_or_symbols:
            i = assure_order_book_id(item)
            current_universe.discard(i)
    else:
        raise RQInvalidArgument(_("unsupported order_book_id type"))

    Environment.get_instance().update_universe(current_universe)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('date').is_valid_date(ignore_none=True),
             verify_that('tenor').is_in(names.VALID_TENORS, ignore_none=True))
def get_yield_curve(date=None, tenor=None):
    """
    获取某个国家市场指定日期的收益率曲线水平。

    数据为2002年至今的中债国债收益率曲线，来源于中央国债登记结算有限责任公司。

    :param date: 查询日期，默认为策略当前日期前一天
    :type date: `str` | `date` | `datetime` | `pandas.Timestamp`

    :param str tenor: 标准期限，'0S' - 隔夜，'1M' - 1个月，'1Y' - 1年，默认为全部期限

    :return: `pandas.DataFrame` - 查询时间段内无风险收益率曲线

    :example:

    ..  code-block:: python3
        :linenos:

        [In]
        get_yield_curve('20130104')

        [Out]
                        0S      1M      2M      3M      6M      9M      1Y      2Y  \
        2013-01-04  0.0196  0.0253  0.0288  0.0279  0.0280  0.0283  0.0292  0.0310

                        3Y      4Y   ...        6Y      7Y      8Y      9Y     10Y  \
        2013-01-04  0.0314  0.0318   ...    0.0342  0.0350  0.0353  0.0357  0.0361
        ...
    """
    data_proxy = ExecutionContext.data_proxy
    dt = ExecutionContext.get_current_trading_dt().date()

    yesterday = data_proxy.get_previous_trading_date(dt)

    if date is None:
        date = yesterday
    else:
        date = pd.Timestamp(date)
        if date > yesterday:
            raise RQInvalidArgument('get_yield_curve: {} >= now({})'.format(date, yesterday))

    return data_proxy.get_yield_curve(start_date=date, end_date=date, tenor=tenor)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('order_book_id').is_valid_instrument(),
             verify_that('bar_count').is_instance_of(int).is_greater_than(0),
             verify_that('frequency').is_in(('1m', '1d')),
             verify_that('fields').are_valid_fields(names.VALID_HISTORY_FIELDS, ignore_none=True),
             verify_that('skip_suspended').is_instance_of(bool))
def history_bars(order_book_id, bar_count, frequency, fields=None, skip_suspended=True):
    """
    获取指定合约的历史行情，同时支持日以及分钟历史数据。不能在init中调用。 注意，该API会自动跳过停牌数据。

    日回测获取分钟历史数据：不支持

    日回测获取日历史数据

    =========================   ===================================================
    调用时间                      返回数据
    =========================   ===================================================
    T日before_trading            T-1日day bar
    T日handle_bar                T日day bar
    =========================   ===================================================

    分钟回测获取日历史数据

    =========================   ===================================================
    调用时间                      返回数据
    =========================   ===================================================
    T日before_trading            T-1日day bar
    T日handle_bar                T-1日day bar
    =========================   ===================================================

    分钟回测获取分钟历史数据

    =========================   ===================================================
    调用时间                      返回数据
    =========================   ===================================================
    T日before_trading            T-1日最后一个minute bar
    T日handle_bar                T日当前minute bar
    =========================   ===================================================

    :param order_book_id: 合约代码或者合约代码列表
    :type order_book_id: `str`

    :param int bar_count: 获取的历史数据数量，必填项

    :param str frequency: 获取数据什么样的频率进行。'1d'或'1m'分别表示每日和每分钟，必填项

    :param str fields: 返回数据字段。必填项。见下方列表。

    =========================   ===================================================
    fields                      字段名
    =========================   ===================================================
    datetime                    时间戳
    open                        开盘价
    high                        最高价
    low                         最低价
    close                       收盘价
    volume                      成交量
    total_turnover              成交额
    datetime                    int类型时间戳
    open_interest               持仓量（期货专用）
    basis_spread                期现差（股指期货专用）
    settlement                  结算价（期货日线专用）
    prev_settlement             结算价（期货日线专用）
    =========================   ===================================================

    :return: `ndarray`, 方便直接与talib等计算库对接，效率较history返回的DataFrame更高。

    :example:

    获取最近5天的日线收盘价序列（策略当前日期为20160706）:

    ..  code-block:: python3
        :linenos:

        [In]
        logger.info(history_bars('000002.XSHE', 5, '1d', 'close'))
        [Out]
        [ 8.69  8.7   8.71  8.81  8.81]
    """
    order_book_id = assure_order_book_id(order_book_id)
    data_proxy = ExecutionContext.data_proxy
    dt = ExecutionContext.get_current_calendar_dt()

    if frequency == '1m' and Environment.get_instance().config.base.frequency == '1d':
        raise RQInvalidArgument('can not get minute history in day back test')

    if (Environment.get_instance().config.base.frequency == '1m' and frequency == '1d') or \
            (frequency == '1d' and ExecutionContext.get_active().phase == EXECUTION_PHASE.BEFORE_TRADING):
        # 在分钟回测获取日线数据, 应该推前一天，这里应该使用 trading date
        dt = data_proxy.get_previous_trading_date(ExecutionContext.get_current_trading_dt().date())

    return data_proxy.history_bars(order_book_id, bar_count, frequency, fields, dt, skip_suspended)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('type').is_in(names.VALID_INSTRUMENT_TYPES, ignore_none=True))
def all_instruments(type=None):
    """
    获取某个国家市场的所有合约信息。使用者可以通过这一方法很快地对合约信息有一个快速了解，目前仅支持中国市场。

    :param str type: 需要查询合约类型，例如：type='CS'代表股票。默认是所有类型

    :return: `pandas DataFrame` 所有合约的基本信息。

    其中type参数传入的合约类型和对应的解释如下：

    =========================   ===================================================
    合约类型                      说明
    =========================   ===================================================
    CS                          Common Stock, 即股票
    ETF                         Exchange Traded Fund, 即交易所交易基金
    LOF                         Listed Open-Ended Fund，即上市型开放式基金
    FenjiMu                     Fenji Mu Fund, 即分级母基金
    FenjiA                      Fenji A Fund, 即分级A类基金
    FenjiB                      Fenji B Funds, 即分级B类基金
    INDX                        Index, 即指数
    Future                      Futures，即期货，包含股指、国债和商品期货
    hour                        int - option [1,4]
    minute                      int - option [1,240]
    =========================   ===================================================

    :example:

    获取中国市场所有分级基金的基础信息:

    ..  code-block:: python3
        :linenos:

        [In]all_instruments('FenjiA')
        [Out]
            abbrev_symbol    order_book_id    product    sector_code  symbol
        0    CYGA    150303.XSHE    null    null    华安创业板50A
        1    JY500A    150088.XSHE    null    null    金鹰500A
        2    TD500A    150053.XSHE    null    null    泰达稳健
        3    HS500A    150110.XSHE    null    null    华商500A
        4    QSAJ    150235.XSHE    null    null    鹏华证券A
        ...

    """
    return ExecutionContext.data_proxy.all_instruments(type)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_symbols').is_instance_of((str, Iterable)))
def instruments(id_or_symbols):
    """
    获取某个国家市场内一个或多个合约的详细信息。目前仅支持中国市场。

    :param order_book_id: 合约代码或者合约代码列表
    :type order_book_id: `str` | List[`str`]

    :return: :class:`~StockInstrument` | :class:`~FutureInstrument`

    目前系统并不支持跨国家市场的同时调用。传入 order_book_id list必须属于同一国家市场，不能混合着中美两个国家市场的order_book_id。

    :example:

    *   获取单一股票合约的详细信息:

        ..  code-block:: python3
            :linenos:

            [In]instruments('000001.XSHE')
            [Out]
            Instrument(order_book_id=000001.XSHE, symbol=平安银行, abbrev_symbol=PAYH, listed_date=19910403, de_listed_date=null, board_type=MainBoard, sector_code_name=金融, sector_code=Financials, round_lot=100, exchange=XSHE, special_type=Normal, status=Active)

    *   获取多个股票合约的详细信息:

        ..  code-block:: python3
            :linenos:

            [In]instruments(['000001.XSHE', '000024.XSHE'])
            [Out]
            [Instrument(order_book_id=000001.XSHE, symbol=平安银行, abbrev_symbol=PAYH, listed_date=19910403, de_listed_date=null, board_type=MainBoard, sector_code_name=金融, sector_code=Financials, round_lot=100, exchange=XSHE, special_type=Normal, status=Active), Instrument(order_book_id=000024.XSHE, symbol=招商地产, abbrev_symbol=ZSDC, listed_date=19930607, de_listed_date=null, board_type=MainBoard, sector_code_name=金融, sector_code=Financials, round_lot=100, exchange=XSHE, special_type=Normal, status=Active)]

    *   获取合约已上市天数:

        ..  code-block:: python
            :linenos:

            instruments('000001.XSHE').days_from_listed()

    *   获取合约距离到期天数:

        ..  code-block:: python
            :linenos:

            instruments('IF1701').days_to_expire()
    """
    return ExecutionContext.data_proxy.instruments(id_or_symbols)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('code').is_instance_of((str, SectorCodeItem)))
def sector(code):
    if not isinstance(code, six.string_types):
        code = code.name
    else:
        code = to_sector_name(code)

    return ExecutionContext.data_proxy.sector(code)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('code').is_instance_of((str, IndustryCodeItem)))
def industry(code):
    if not isinstance(code, six.string_types):
        code = code.code
    else:
        code = to_industry_code(code)

    return ExecutionContext.data_proxy.industry(code)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
def concept(*concept_names):
    return ExecutionContext.data_proxy.concept(*concept_names)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('start_date').is_valid_date(ignore_none=False))
@apply_rules(verify_that('end_date').is_valid_date(ignore_none=False))
def get_trading_dates(start_date, end_date):
    """
    获取某个国家市场的交易日列表（起止日期加入判断）。目前仅支持中国市场。

    :param start_date: 开始日期
    :type start_date: `str` | `date` | `datetime` | `pandas.Timestamp`

    :param end_date: 结束如期
    :type end_date: `str` | `date` | `datetime` | `pandas.Timestamp`

    :return: list[`datetime.date`]

    :example:

    ..  code-block:: python3
        :linenos:

        [In]get_trading_dates(start_date='2016-05-05', end_date='20160505')
        [Out]
        [datetime.date(2016, 5, 5)]
    """
    return ExecutionContext.data_proxy.get_trading_dates(start_date, end_date)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('date').is_valid_date(ignore_none=False))
def get_previous_trading_date(date):
    """
    获取指定日期的上一交易日。

    :param date: 指定日期
    :type date: `str` | `date` | `datetime` | `pandas.Timestamp`

    :return: `datetime.date`

    :example:

    ..  code-block:: python3
        :linenos:

        [In]get_previous_trading_date(date='2016-05-02')
        [Out]
        [datetime.date(2016, 4, 29)]
    """
    return ExecutionContext.data_proxy.get_previous_trading_date(date)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('date').is_valid_date(ignore_none=False))
def get_next_trading_date(date):
    """
    获取指定日期的下一交易日

    :param date: 指定日期
    :type date: `str` | `date` | `datetime` | `pandas.Timestamp`

    :return: `datetime.date`

    :example:

    ..  code-block:: python3
        :linenos:

        [In]get_next_trading_date(date='2016-05-01')
        [Out]
        [datetime.date(2016, 5, 3)]
    """
    return ExecutionContext.data_proxy.get_next_trading_date(date)


def to_date(date):
    if isinstance(date, six.string_types):
        return parse(date).date()

    if isinstance(date, datetime.date):
        try:
            return date.date()
        except AttributeError:
            return date
    
    raise RQInvalidArgument('unknown date value: {}'.format(date))


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_INIT,
                                EXECUTION_PHASE.BEFORE_TRADING,
                                EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.AFTER_TRADING,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('order_book_id').is_valid_instrument(),
             verify_that('start_date').is_valid_date(ignore_none=False),
             verify_that('adjusted').is_instance_of(bool))
def get_dividend(order_book_id, start_date, adjusted=True):
    dt = ExecutionContext.get_current_trading_dt().date() - datetime.timedelta(days=1)
    start_date = to_date(start_date)
    if start_date > dt:
        raise RQInvalidArgument(
            _('in get_dividend, start_date {} is later than the previous test day {}').format(
                start_date, dt
            ))
    order_book_id = assure_order_book_id(order_book_id)
    df = ExecutionContext.data_proxy.get_dividend(order_book_id, adjusted)
    return df[start_date:dt]


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('series_name').is_instance_of(str),
             verify_that('value').is_number())
def plot(series_name, value):
    """
    Add a point to custom series.
    :param str series_name: the name of custom series
    :param float value: the value of the series in this time
    :return: None
    """
    if ExecutionContext.plots is None:
        # FIXME: this is ugly
        from ..utils.plot_store import PlotStore
        ExecutionContext.plots = PlotStore()
    ExecutionContext.plots.add_plot(ExecutionContext.trading_dt.date(), series_name, value)


@export_as_api
@ExecutionContext.enforce_phase(EXECUTION_PHASE.ON_BAR,
                                EXECUTION_PHASE.ON_TICK,
                                EXECUTION_PHASE.SCHEDULED)
@apply_rules(verify_that('id_or_symbol').is_valid_instrument())
def current_snapshot(id_or_symbol):
    """
    获得当前市场快照数据。只能在日内交易阶段调用，获取当日调用时点的市场快照数据。市场快照数据记录了每日从开盘到当前的数据信息，可以理解为一个动态的day bar数据。在目前分钟回测中，快照数据为当日所有分钟线累积而成，一般情况下，最后一个分钟线获取到的快照数据应当与当日的日线行情保持一致。需要注意，在实盘模拟中，该函数返回的是调用当时的市场快照情况，所以在同一个handle_bar中不同时点调用可能返回的数据不同。如果当日截止到调用时候对应股票没有任何成交，那么snapshot中的close, high, low, last几个价格水平都将以0表示。

    :param str order_book_id: 合约代码或简称

    :return: :class:`~Snapshot`

    :example:

    在handle_bar中调用该函数，假设策略当前时间是20160104 09:33:

    ..  code-block:: python3
        :linenos:

        [In]
        logger.info(current_snapshot('000001.XSHE'))
        [Out]
        2016-01-04 09:33:00.00  INFO
        Snapshot(order_book_id: '000001.XSHE', datetime: datetime.datetime(2016, 1, 4, 9, 33), open: 10.0, high: 10.025, low: 9.9667, last: 9.9917, volume: 2050320, total_turnover: 20485195, prev_close: 9.99)
    """
    data_proxy = ExecutionContext.data_proxy
    cal_dt = ExecutionContext.get_current_calendar_dt()

    frequency = Environment.get_instance().config.base.frequency

    order_book_id = assure_order_book_id(id_or_symbol)
    return data_proxy.current_snapshot(order_book_id, frequency, cal_dt)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base_position import BasePosition, PositionClone
from .future_position import FuturePosition
from .stock_position import StockPosition


class Positions(dict):
    def __init__(self, position_type):
        super(Positions, self).__init__()
        self._position_type = position_type

    def __missing__(self, key):
        p = self._position_type(key)
        self[key] = p
        return p

    def _clone(self):
        ps = {}
        for order_book_id in self:
            ps[order_book_id] = self[order_book_id]._clone()
        return ps


class PositionsClone(dict):
    def __init__(self, position_type):
        super(PositionsClone, self).__init__()
        self._position_type = position_type

    def __missing__(self, key):
        p = PositionClone()
        position = self._position_type(key)
        for key in dir(position):
            if "__" in key:
                continue
            setattr(p, key, getattr(position, key))
        return p

    def __repr__(self):
        return str([order_book_id for order_book_id in self])

    def __iter__(self):
        for key in sorted(super(PositionsClone, self).keys()):
            yield key

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def keys(self):
        for key in sorted(super(PositionsClone, self).keys()):
            yield key



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
try:
    # For Python 2 兼容
    from functools import lru_cache
except Exception as e:
    from fastcache import lru_cache
import datetime


TimeRange = namedtuple('TimeRange', ['start', 'end'])


def get_month_begin_time(time=None):
    if time is None:
        time = datetime.datetime.now()
    return time.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def get_month_end_time(time=None):
    try:
        return time.replace(month=time.month + 1, day=1, hour=23, minute=59, second=59,
                            microsecond=999) - datetime.timedelta(days=1)
    except ValueError:
        return time.replace(year=time.year + 1, month=1, day=1, hour=23, minute=59, second=59,
                            microsecond=999) - datetime.timedelta(days=1)


def get_last_date(trading_calendar, dt):
    idx = trading_calendar.searchsorted(dt)
    return trading_calendar[idx - 1]


def convert_date_to_int(dt):
    t = dt.year * 10000 + dt.month * 100 + dt.day
    t *= 1000000
    return t


def convert_dt_to_int(dt):
    t = convert_date_to_int(dt)
    t += dt.hour * 10000 + dt.minute * 100 + dt.second
    return t


def convert_int_to_date(dt_int):
    dt_int = int(dt_int)
    if dt_int > 100000000:
        dt_int //= 1000000
    return _convert_int_to_date(dt_int)


@lru_cache(None)
def _convert_int_to_date(dt_int):
    year, r = divmod(dt_int, 10000)
    month, day = divmod(r, 100)
    return datetime.datetime(year, month, day)


@lru_cache(20480)
def convert_int_to_datetime(dt_int):
    dt_int = int(dt_int)
    year, r = divmod(dt_int, 10000000000)
    month, r = divmod(r, 100000000)
    day, r = divmod(r, 1000000)
    hour, r = divmod(r, 10000)
    minute, second = divmod(r, 100)
    return datetime.datetime(year, month, day, hour, minute, second)




# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from ..execution_context import ExecutionContext
from .api_base import instruments


def get_current_bar_dict():
    bar_dict = ExecutionContext.get_current_bar_dict()
    return bar_dict


def price_change(stock):
    bar_dict = get_current_bar_dict()
    return bar_dict[stock].close / bar_dict[stock].prev_close - 1


def symbol(order_book_id, split=", "):
    if isinstance(order_book_id, six.string_types):
        return "{}[{}]".format(order_book_id, instruments(order_book_id).symbol)
    else:
        s = split.join(symbol(item) for item in order_book_id)
        return s


def now_time_str(str_format="%H:%M:%S"):
    dt = ExecutionContext.get_current_trading_dt()
    return dt.strftime(str_format)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

DividendPersistMap = {
    "order_book_id": "order_book_id",
    "quantity": "quantity",
    "dividend_series_dict": "dividend_series_dict",
}


class Dividend(object):
    def __init__(self, order_book_id, quantity, dividend_series_dict):
        self.order_book_id = order_book_id
        self.quantity = quantity
        self.dividend_series_dict = dividend_series_dict

    @classmethod
    def __from_dict__(cls, dividend_dict):
        return cls(dividend_dict['order_book_id'], dividend_dict["quantity"], dividend_dict["dividend_series_dict"])

    def __to_dict__(self):
        dividend_dict = {}
        for persist_key, origin_key in six.iteritems(DividendPersistMap):
            dividend_dict[persist_key] = getattr(self, origin_key)
        return dividend_dict



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from contextlib import contextmanager

from .utils.i18n import gettext as _
from .utils.exception import CustomException, patch_user_exc
from .utils import get_upper_underlying_symbol
from .utils.default_future_info import DEFAULT_FUTURE_INFO


class ContextStack(object):
    def __init__(self):
        self.stack = []

    def push(self, obj):
        self.stack.append(obj)

    def pop(self):
        try:
            return self.stack.pop()
        except IndexError:
            raise RuntimeError("stack is empty")

    @contextmanager
    def pushed(self, obj):
        self.push(obj)
        try:
            yield self
        finally:
            self.pop()

    @property
    def top(self):
        try:
            return self.stack[-1]
        except IndexError:
            raise RuntimeError("stack is empty")


class ExecutionContext(object):
    stack = ContextStack()
    config = None
    data_proxy = None
    account = None
    accounts = None
    broker = None
    calendar_dt = None
    trading_dt = None
    plots = None

    def __init__(self, phase, bar_dict=None):
        self.phase = phase
        self.bar_dict = bar_dict

    def _push(self):
        self.stack.push(self)

    def _pop(self):
        popped = self.stack.pop()
        if popped is not self:
            raise RuntimeError("Popped wrong context")
        return self

    def __enter__(self):
        self._push()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Restore the algo instance stored in __enter__.
        """
        if exc_type is None:
            self._pop()
            return False

        # 处理嵌套ExecutionContext
        last_exc_val = exc_val
        while isinstance(exc_val, CustomException):
            last_exc_val = exc_val
            if exc_val.error.exc_val is not None:
                exc_val = exc_val.error.exc_val
            else:
                break
        if isinstance(last_exc_val, CustomException):
            raise last_exc_val

        from .utils import create_custom_exception
        user_exc = create_custom_exception(exc_type, exc_val, exc_tb, self.config.base.strategy_file)
        raise user_exc

    @classmethod
    def get_active(cls):
        return cls.stack.top

    @classmethod
    def get_current_bar_dict(cls):
        ctx = cls.get_active()
        return ctx.bar_dict

    @classmethod
    def get_current_calendar_dt(cls):
        return ExecutionContext.calendar_dt

    @classmethod
    def get_current_trading_dt(cls):
        return ExecutionContext.trading_dt

    @classmethod
    def get_current_run_id(cls):
        return ExecutionContext.config.base.run_id

    @classmethod
    def get_instrument(cls, order_book_id):
        return ExecutionContext.data_proxy.instruments(order_book_id)

    @classmethod
    def get_data_proxy(cls):
        return ExecutionContext.data_proxy

    @classmethod
    def enforce_phase(cls, *phases):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                phase = cls.get_active().phase
                if phase not in phases:
                    raise patch_user_exc(
                        RuntimeError(_("You cannot call %s when executing %s") % (func.__name__, phase.value)))
                return func(*args, **kwargs)
            return wrapper
        return decorator

    @classmethod
    def get_current_close_price(cls, order_book_id):
        return ExecutionContext.data_proxy.current_snapshot(
            order_book_id,
            ExecutionContext.config.base.frequency,
            ExecutionContext.calendar_dt
        ).last

    @classmethod
    def get_future_commission_info(cls, order_book_id, hedge_type):
        try:
            return ExecutionContext.data_proxy.get_future_info(order_book_id, hedge_type)
        except NotImplementedError:
            underlying_symbol = get_upper_underlying_symbol(order_book_id)
            return DEFAULT_FUTURE_INFO[underlying_symbol][hedge_type.value]

    @classmethod
    def get_future_margin(cls, order_book_id):
        try:
            return ExecutionContext.data_proxy.get_future_info(order_book_id)['long_margin_ratio']
        except NotImplementedError:
            return ExecutionContext.data_proxy.instruments(order_book_id).margin_rate

    @classmethod
    def get_future_info(cls, order_book_id, hedge_type):
        return ExecutionContext.data_proxy.get_future_info(order_book_id, hedge_type)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import datetime

from ..execution_context import ExecutionContext
from ..utils import instrument_type_str2enum


class Instrument(object):
    DEFAULT_LISTED_DATE = datetime.datetime(1990, 1, 1)
    DEFAULT_DE_LISTED_DATE = datetime.datetime(2999, 12, 31)

    @staticmethod
    def _fix_date(ds, dflt):
        if ds == '0000-00-00':
            return dflt
        year, month, day = ds.split('-')
        return datetime.datetime(int(year), int(month), int(day))

    def __init__(self, dic):
        self.__dict__ = dic
        if 'listed_date' in self.__dict__:
            self.listed_date = self._fix_date(self.listed_date, self.DEFAULT_LISTED_DATE)
        if 'de_listed_date' in self.__dict__:
            self.de_listed_date = self._fix_date(self.de_listed_date, self.DEFAULT_DE_LISTED_DATE)
        if 'maturity_date' in self.__dict__:
            self.maturity_date = self._fix_date(self.maturity_date, self.DEFAULT_DE_LISTED_DATE)

    def __repr__(self):
        return "{}({})".format(type(self).__name__,
                               ", ".join(["{}={}".format(k, repr(v))
                                          for k, v in six.iteritems(self.__dict__)]))

    @property
    def listing(self):
        now = ExecutionContext.get_current_calendar_dt()
        return self.listed_date <= now <= self.de_listed_date

    def days_from_listed(self):
        if self.listed_date == self.DEFAULT_LISTED_DATE:
            return -1

        date = ExecutionContext.get_current_trading_dt().date()
        if self.de_listed_date.date() < date:
            return -1

        ipo_days = (date - self.listed_date.date()).days
        return ipo_days if ipo_days >= 0 else -1

    @property
    def enum_type(self):
        return instrument_type_str2enum(self.type)

    def days_to_expire(self):
        if self.type != 'Future' or self.order_book_id[-2:] == '88' or self.order_book_id[-2:] == '99':
            return -1
        
        date = ExecutionContext.get_current_trading_dt().date()
        days = (self.maturity_date.date() - date).days
        return -1 if days < 0 else days


class SectorCodeItem(object):
    def __init__(self, cn, en, name):
        self.__cn = cn
        self.__en = en
        self.__name = name

    @property
    def cn(self):
        return self.__cn

    @property
    def en(self):
        return self.__en

    @property
    def name(self):
        return self.__name

    def __repr__(self):
        return "{}: {}, {}".format(self.__name, self.__en, self.__cn)


class SectorCode(object):
    Energy = SectorCodeItem("能源", "energy", 'Energy')
    Materials = SectorCodeItem("原材料", "materials", 'Materials')
    ConsumerDiscretionary = SectorCodeItem("非必需消费品", "consumer discretionary", 'ConsumerDiscretionary')
    ConsumerStaples = SectorCodeItem("必需消费品", "consumer staples", 'ConsumerStaples')
    HealthCare = SectorCodeItem("医疗保健", "health care", 'HealthCare')
    Financials = SectorCodeItem("金融", "financials", 'Financials')
    InformationTechnology = SectorCodeItem("信息技术", "information technology", 'InformationTechnology')
    TelecommunicationServices = SectorCodeItem("电信服务", "telecommunication services", 'TelecommunicationServices')
    Utilities = SectorCodeItem("公共服务", "utilities", 'Utilities')
    Industrials = SectorCodeItem("工业", "industrials", "Industrials")


class IndustryCodeItem(object):
    def __init__(self, code, name):
        self.__code = code
        self.__name = name

    @property
    def code(self):
        return self.__code

    @property
    def name(self):
        return self.__name

    def __repr__(self):
        return "{0}:{1}".format(self.__code,self.__name)


class IndustryCode(object):
    A01 = IndustryCodeItem("A01", "农业")
    A02 = IndustryCodeItem("A02", "林业")
    A03 = IndustryCodeItem("A03", "畜牧业")
    A04 = IndustryCodeItem("A04", "渔业")
    A05 = IndustryCodeItem("A05", "农、林、牧、渔服务业")
    B06 = IndustryCodeItem("B06", "煤炭开采和洗选业")
    B07 = IndustryCodeItem("B07", "石油和天然气开采业")
    B08 = IndustryCodeItem("B08", "黑色金属矿采选业")
    B09 = IndustryCodeItem("B09", "有色金属矿采选业")
    B10 = IndustryCodeItem("B10", "非金属矿采选业")
    B11 = IndustryCodeItem("B11", "开采辅助活动")
    B12 = IndustryCodeItem("B12", "其他采矿业")
    C13 = IndustryCodeItem("C13", "农副食品加工业")
    C14 = IndustryCodeItem("C14", "食品制造业")
    C15 = IndustryCodeItem("C15", "酒、饮料和精制茶制造业")
    C16 = IndustryCodeItem("C16", "烟草制品业")
    C17 = IndustryCodeItem("C17", "纺织业")
    C18 = IndustryCodeItem("C18", "纺织服装、服饰业")
    C19 = IndustryCodeItem("C19", "皮革、毛皮、羽毛及其制品和制鞋业")
    C20 = IndustryCodeItem("C20", "木材加工及木、竹、藤、棕、草制品业")
    C21 = IndustryCodeItem("C21", "家具制造业")
    C22 = IndustryCodeItem("C22", "造纸及纸制品业")
    C23 = IndustryCodeItem("C23", "印刷和记录媒介复制业")
    C24 = IndustryCodeItem("C24", "文教、工美、体育和娱乐用品制造业")
    C25 = IndustryCodeItem("C25", "石油加工、炼焦及核燃料加工业")
    C26 = IndustryCodeItem("C26", "化学原料及化学制品制造业")
    C27 = IndustryCodeItem("C27", "医药制造业")
    C28 = IndustryCodeItem("C28", "化学纤维制造业")
    C29 = IndustryCodeItem("C29", "橡胶和塑料制品业")
    C30 = IndustryCodeItem("C30", "非金属矿物制品业")
    C31 = IndustryCodeItem("C31", "黑色金属冶炼及压延加工业")
    C32 = IndustryCodeItem("C32", "有色金属冶炼和压延加工业")
    C33 = IndustryCodeItem("C33", "金属制品业")
    C34 = IndustryCodeItem("C34", "通用设备制造业")
    C35 = IndustryCodeItem("C35", "专用设备制造业")
    C36 = IndustryCodeItem("C36", "汽车制造业")
    C37 = IndustryCodeItem("C37", "铁路、船舶、航空航天和其它运输设备制造业")
    C38 = IndustryCodeItem("C38", "电气机械及器材制造业")
    C39 = IndustryCodeItem("C39", "计算机、通信和其他电子设备制造业")
    C40 = IndustryCodeItem("C40", "仪器仪表制造业")
    C41 = IndustryCodeItem("C41", "其他制造业")
    C42 = IndustryCodeItem("C42", "废弃资源综合利用业")
    C43 = IndustryCodeItem("C43", "金属制品、机械和设备修理业")
    D44 = IndustryCodeItem("D44", "电力、热力生产和供应业")
    D45 = IndustryCodeItem("D45", "燃气生产和供应业")
    D46 = IndustryCodeItem("D46", "水的生产和供应业")
    E47 = IndustryCodeItem("E47", "房屋建筑业")
    E48 = IndustryCodeItem("E48", "土木工程建筑业")
    E49 = IndustryCodeItem("E49", "建筑安装业")
    E50 = IndustryCodeItem("E50", "建筑装饰和其他建筑业")
    F51 = IndustryCodeItem("F51", "批发业")
    F52 = IndustryCodeItem("F52", "零售业")
    G53 = IndustryCodeItem("G53", "铁路运输业")
    G54 = IndustryCodeItem("G54", "道路运输业")
    G55 = IndustryCodeItem("G55", "水上运输业")
    G56 = IndustryCodeItem("G56", "航空运输业")
    G57 = IndustryCodeItem("G57", "管道运输业")
    G58 = IndustryCodeItem("G58", "装卸搬运和运输代理业")
    G59 = IndustryCodeItem("G59", "仓储业")
    G60 = IndustryCodeItem("G60", "邮政业")
    H61 = IndustryCodeItem("H61", "住宿业")
    H62 = IndustryCodeItem("H62", "餐饮业")
    I63 = IndustryCodeItem("I63", "电信、广播电视和卫星传输服务")
    I64 = IndustryCodeItem("I64", "互联网和相关服务")
    I65 = IndustryCodeItem("I65", "软件和信息技术服务业")
    J66 = IndustryCodeItem("J66", "货币金融服务")
    J67 = IndustryCodeItem("J67", "资本市场服务")
    J68 = IndustryCodeItem("J68", "保险业")
    J69 = IndustryCodeItem("J69", "其他金融业")
    K70 = IndustryCodeItem("K70", "房地产业")
    L71 = IndustryCodeItem("L71", "租赁业")
    L72 = IndustryCodeItem("L72", "商务服务业")
    M73 = IndustryCodeItem("M73", "研究和试验发展")
    M74 = IndustryCodeItem("M74", "专业技术服务业")
    M75 = IndustryCodeItem("M75", "科技推广和应用服务业")
    N76 = IndustryCodeItem("N76", "水利管理业")
    N77 = IndustryCodeItem("N77", "生态保护和环境治理业")
    N78 = IndustryCodeItem("N78", "公共设施管理业")
    O79 = IndustryCodeItem("O79", "居民服务业")
    O80 = IndustryCodeItem("O80", "机动车、电子产品和日用产品修理业")
    O81 = IndustryCodeItem("O81", "其他服务业")
    P82 = IndustryCodeItem("P82", "教育")
    Q83 = IndustryCodeItem("Q83", "卫生")
    Q84 = IndustryCodeItem("Q84", "社会工作")
    R85 = IndustryCodeItem("R85", "新闻和出版业")
    R86 = IndustryCodeItem("R86", "广播、电视、电影和影视录音制作业")
    R87 = IndustryCodeItem("R87", "文化艺术业")
    R88 = IndustryCodeItem("R88", "体育")
    R89 = IndustryCodeItem("R89", "娱乐业")
    S90 = IndustryCodeItem("S90", "综合")



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from rqalpha.utils.i18n import gettext as _
from rqalpha.const import ORDER_TYPE, SIDE, BAR_STATUS
from rqalpha.model.trade import Trade
from rqalpha.environment import Environment
from rqalpha.events import EVENT


class Matcher(object):
    def __init__(self,
                 deal_price_decider,
                 bar_limit=True,
                 volume_percent=0.25):
        self._board = None
        self._turnover = defaultdict(int)
        self._calendar_dt = None
        self._trading_dt = None
        self._deal_price_decider = deal_price_decider
        self._volume_percent = volume_percent
        self._bar_limit = bar_limit

    def update(self, calendar_dt, trading_dt, bar_dict):
        self._board = bar_dict
        self._turnover.clear()
        self._calendar_dt = calendar_dt
        self._trading_dt = trading_dt

    def match(self, open_orders):
        for account, order in open_orders:
            slippage_decider = account.slippage_decider
            commission_decider = account.commission_decider
            tax_decider = account.tax_decider

            bar = self._board[order.order_book_id]
            bar_status = bar._bar_status

            if bar_status == BAR_STATUS.ERROR:
                listed_date = bar.instrument.listed_date.date()
                if listed_date == self._trading_dt.date():
                    reason = _("Order Cancelled: current security [{order_book_id}] can not be traded in listed date [{listed_date}]").format(
                        order_book_id=order.order_book_id,
                        listed_date=listed_date,
                    )
                else:
                    reason = _("Order Cancelled: current bar [{order_book_id}] miss market data.").format(
                        order_book_id=order.order_book_id)
                order._mark_rejected(reason)
                continue

            deal_price = self._deal_price_decider(bar)
            if order.type == ORDER_TYPE.LIMIT:
                if order.price > bar.limit_up:
                    reason = _(
                        "Order Rejected: limit order price {limit_price} is higher than limit up {limit_up}."
                    ).format(
                        limit_price=order.price,
                        limit_up=bar.limit_up
                    )
                    order._mark_rejected(reason)
                    continue

                if order.price < bar.limit_down:
                    reason = _(
                        "Order Rejected: limit order price {limit_price} is lower than limit down {limit_down}."
                    ).format(
                        limit_price=order.price,
                        limit_down=bar.limit_down
                    )
                    order._mark_rejected(reason)
                    continue

                if order.side == SIDE.BUY and order.price < deal_price:
                    continue
                if order.side == SIDE.SELL and order.price > deal_price:
                    continue
            else:
                if self._bar_limit and order.side == SIDE.BUY and bar_status == BAR_STATUS.LIMIT_UP:
                    reason = _(
                        "Order Cancelled: current bar [{order_book_id}] reach the limit_up price."
                    ).format(order_book_id=order.order_book_id)
                    order._mark_rejected(reason)
                    continue
                elif self._bar_limit and order.side == SIDE.SELL and bar_status == BAR_STATUS.LIMIT_DOWN:
                    reason = _(
                        "Order Cancelled: current bar [{order_book_id}] reach the limit_down price."
                    ).format(order_book_id=order.order_book_id)
                    order._mark_rejected(reason)
                    continue

            if self._bar_limit:
                if order.side == SIDE.BUY and bar_status == BAR_STATUS.LIMIT_UP:
                    continue
                if order.side == SIDE.SELL and bar_status == BAR_STATUS.LIMIT_DOWN:
                    continue

            volume_limit = round(bar.volume * self._volume_percent) - self._turnover[order.order_book_id]
            round_lot = bar.instrument.round_lot
            volume_limit = (volume_limit // round_lot) * round_lot
            if volume_limit <= 0:
                if order.type == ORDER_TYPE.MARKET:
                    reason = _('Order Cancelled: market order {order_book_id} volume {order_volume}'
                               ' due to volume limit').format(
                        order_book_id=order.order_book_id,
                        order_volume=order.quantity
                    )
                    order._mark_cancelled(reason)
                continue

            unfilled = order.unfilled_quantity
            fill = min(unfilled, volume_limit)
            ct_amount = account.portfolio.positions[order.order_book_id]._cal_close_today_amount(fill, order.side)
            price = slippage_decider.get_trade_price(order, deal_price)
            trade = Trade.__from_create__(order=order, calendar_dt=self._calendar_dt, trading_dt=self._trading_dt,
                                          price=price, amount=fill, close_today_amount=ct_amount)
            trade._commission = commission_decider.get_commission(trade)
            trade._tax = tax_decider.get_tax(trade)
            order._fill(trade)
            self._turnover[order.order_book_id] += fill

            Environment.get_instance().event_bus.publish_event(EVENT.TRADE, account, trade)

            if order.type == ORDER_TYPE.MARKET and order.unfilled_quantity != 0:
                reason = _(
                    "Order Cancelled: market order {order_book_id} volume {order_volume} is"
                    " larger than 25 percent of current bar volume, fill {filled_volume} actually"
                ).format(
                    order_book_id=order.order_book_id,
                    order_volume=order.quantity,
                    filled_volume=order.filled_quantity
                )
                order._mark_cancelled(reason)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Ricequant, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from six import with_metaclass
from collections import defaultdict

from ..utils import get_upper_underlying_symbol
from ..utils.default_future_info import DEFAULT_FUTURE_INFO
from ..const import ACCOUNT_TYPE, HEDGE_TYPE, SIDE, COMMISSION_TYPE, POSITION_EFFECT
from ..execution_context import ExecutionContext


def init_commission(account_type, multiplier):
    if account_type in [ACCOUNT_TYPE.STOCK, ACCOUNT_TYPE.BENCHMARK]:
        return StockCommission(multiplier)
    elif account_type == ACCOUNT_TYPE.FUTURE:
        return FutureCommission(multiplier)
    else:
        raise NotImplementedError


class BaseCommission(with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def get_commission(self, trade):
        raise NotImplementedError


class StockCommission(BaseCommission):
    def __init__(self, multiplier, min_commission=5):
        self.rate = 0.0008
        self.multiplier = multiplier
        self.commission_map = defaultdict(lambda: min_commission)
        self.min_commission = min_commission

    def get_commission(self, trade):
        """
        计算手续费这个逻辑比较复杂，按照如下算法来计算：
        1.  定义一个剩余手续费的概念，根据order_id存储在commission_map中，默认为min_commission
        2.  当trade来时计算该trade产生的手续费cost_money
        3.  如果cost_money > commission
            3.1 如果commission 等于 min_commission，说明这是第一笔trade，此时，直接commission置0，返回cost_money即可
            3.2 如果commission 不等于 min_commission, 则说明这不是第一笔trade,此时，直接cost_money - commission即可
        4.  如果cost_money <= commission
            4.1 如果commission 等于 min_commission, 说明是第一笔trade, 此时，返回min_commission(提前把最小手续费收了)
            4.2 如果commission 不等于 min_commission， 说明不是第一笔trade, 之前的trade中min_commission已经收过了，所以返回0.
        """
        order_id = trade.order.order_id
        commission = self.commission_map[order_id]
        cost_money = trade.last_price * trade.last_quantity * self.rate * self.multiplier
        if cost_money > commission:
            if commission == self.min_commission:
                self.commission_map[order_id] = 0
                return cost_money
            else:
                self.commission_map[order_id] = 0
                return cost_money - commission
        else:
            if commission == self.min_commission:
                self.commission_map[order_id] -= cost_money
                return commission
            else:
                self.commission_map[order_id] -= cost_money
                return 0


class FutureCommission(BaseCommission):
    def __init__(self, multiplier, hedge_type=HEDGE_TYPE.SPECULATION):
        """
        期货目前不计算最小手续费
        """
        self.multiplier = multiplier
        self.hedge_type = hedge_type

    def get_commission(self, trade):
        order = trade.order
        order_book_id = order.order_book_id
        info = ExecutionContext.get_future_commission_info(order_book_id, self.hedge_type)
        commission = 0
        if info['commission_type'] == COMMISSION_TYPE.BY_MONEY:
            contract_multiplier = ExecutionContext.get_instrument(order.order_book_id).contract_multiplier
            if trade.order.position_effect == POSITION_EFFECT.OPEN:
                commission += trade.last_price * trade.last_quantity * contract_multiplier * info['open_commission_ratio']
            else:
                commission += trade.last_price * (trade.last_quantity - trade._close_today_amount) * contract_multiplier * info[
                    'close_commission_ratio']
                commission += trade.last_price * trade._close_today_amount * contract_multiplier * info[
                    'close_commission_today_ratio']
        else:
            if trade.order.position_effect == POSITION_EFFECT.OPEN:
                commission += trade.last_quantity * info['open_commission_ratio']
            else:
                commission += (trade.last_quantity - trade._close_today_amount) * info['close_commission_ratio']
                commission += trade._close_today_amount * info['close_commission_today_ratio']
        return commission * self.multiplier
